// Copyright (c) 2016, Thomas L. Falch
// For conditions of distribution and use, see the accompanying LICENSE and README files

// This file is part of the ImageCL source-to-source compiler
// developed at the Norwegian University of Science and technology


#include "arrayflattener.h"
#include "astutil.h"

using namespace std;
using namespace SageBuilder;
using namespace SageInterface;

ArrayFlattener::ArrayFlattener(SgProject *project, Parameters params, KernelInfo kernelInfo, Settings settings)
        : project(project), params(params), kernelInfo(kernelInfo), settings(settings) {
    arraysTransformed = new set<SgName>();
}

void ArrayFlattener::transform() {
    for (SgNode *varRefNode : NodeQuery::querySubTree(project, V_SgVarRefExp)) {
        SgVarRefExp *varRef = isSgVarRefExp(varRefNode);
        const char *name = varRef->get_symbol()->get_name().str();
        if (kernelInfo.isImageArray(name)) {
            //cout << "[Image]: " << name << endl;
            flatten(varRef);
        }
        if (kernelInfo.isBlockArray(name)) {
            //cout << "[Block]: " << name << endl;
            flatten3d(varRef);
        }
        if (kernelInfo.aliasArrays.count(name)) {
            auto &alias = kernelInfo.aliasArrays[name];
            auto *aliasVarRef = buildVarRefExp(alias.first, varRef->get_symbol()->get_scope());
            flatten3d(varRef, aliasVarRef);
        }
    }

    // Parameters
    SgFunctionDeclaration *funcDec = AstUtil::getFunctionDeclaration(project, kernelInfo.getKernelName());
    transformParameters(funcDec->get_parameterList());

    // Variables
    for (auto n : NodeQuery::querySubTree(project, V_SgVariableDeclaration)) {
        auto *varDecl = isSgVariableDeclaration(n);
        for (auto &a : varDecl->get_variables()) {
            if (AstUtil::isImageType(a->get_type()) || AstUtil::isBlockType(a->get_type())) {
                SgType *baseType;
                switch (kernelInfo.getPixelType(a->get_name().str())) {
                    case FLOAT:
                        baseType = buildFloatType();
                        break;
                    case INT:
                        baseType = buildIntType();
                        break;
                    case UCHAR:
                        baseType = buildUnsignedCharType();
                        break;
                    case DOUBLE:
                        baseType = buildDoubleType();
                        break;
                    default:
                        baseType = buildFloatType();
                        break;
                }
                a->set_type(buildPointerType(baseType));
            }
        }
    }
    AstUtil::fixUniqueNameAttributes(project);
}

void ArrayFlattener::flatten(SgVarRefExp *array, SgVarRefExp *alias) {
    //Here we know its 2D, we'll have a separate function for 3D (and/or 1D)

    SgNode *middleNode = array->get_parent();
    SgPntrArrRefExp *middle = isSgPntrArrRefExp(middleNode);

    //This is if input is argument of read_image etc.
    if (!middle) {
        return;
    }

    SgNode *arrayRefNode = middle->get_parent();
    SgPntrArrRefExp *arrayRef = isSgPntrArrRefExp(arrayRefNode);

    SgExpression *x = middle->get_rhs_operand();
    SgExpression *y = arrayRef->get_rhs_operand();

    SgScopeStatement *scope = getEnclosingScope(array);

    SgExpression *indexCalculation;
    string arrayName = array->get_symbol()->get_name().str();
    if (alias) {
        arrayName = alias->get_symbol()->get_name().str();
    }

    SgExpression *arrayWidth;
    if (settings.generateMPI && kernelInfo.needsMpiScatter(arrayName)) {
        HaloSize hs = kernelInfo.getFootprintTable().at(arrayName).computeHaloSize();
        arrayWidth = buildAddOp(buildVarRefExp(arrayName + "_width", scope), buildIntVal(hs.left + hs.right));
    } else {
        arrayWidth = buildVarRefExp(arrayName + "_width", scope);
    }
    indexCalculation = buildAddOp(buildMultiplyOp(y, arrayWidth), x);

    arrayRef->set_lhs_operand(array);
    arrayRef->set_rhs_operand(indexCalculation);

    arraysTransformed->insert(arrayName);
}

void ArrayFlattener::flatten3d(SgVarRefExp *array, SgVarRefExp *alias) {
    if (auto *middle = isSgPntrArrRefExp(array->get_parent())) {

        auto *parent = isSgPntrArrRefExp(middle->get_parent());
        if (!parent) { exit(401); }
        auto *grand_parent = isSgPntrArrRefExp(parent->get_parent());
        if (!grand_parent) { exit(402); }
        auto *x = middle->get_rhs_operand();
        auto *y = parent->get_rhs_operand();
        auto *z = grand_parent->get_rhs_operand();

        //cout << "x: " << x->unparseToString() << endl;
        //cout << "y: " << y->unparseToString() << endl;
        //cout << "z: " << z->unparseToString() << endl;

        auto *scope = getEnclosingScope(array);
        string array_name = array->get_symbol()->get_name().str();
        if (alias) {
            array_name = alias->get_symbol()->get_name().str();
        }
        SgExpression *array_width;
        SgExpression *array_height;
        if (settings.generateMPI && kernelInfo.needsMpiScatter(array_name)) {
            HaloSize hs = kernelInfo.getFootprintTable().at(array_name).computeHaloSize();
            array_width = buildAddOp(buildVarRefExp(array_name + "_width", scope), buildIntVal(hs.left + hs.right));
            array_height = buildAddOp(buildVarRefExp(array_name + "_height", scope), buildIntVal(hs.up + hs.down));
        } else {
            array_width = buildVarRefExp(array_name + "_width", scope);
            array_height = buildVarRefExp(array_name + "_height", scope);
        }
        auto *y_index_calculation = buildAddOp(buildMultiplyOp(y, array_width), x);
        auto *z_index_calculation = buildAddOp(buildMultiplyOp(z, buildMultiplyOp(array_width, array_height)), y_index_calculation);

        grand_parent->set_lhs_operand(array);
        grand_parent->set_rhs_operand(z_index_calculation);

        arraysTransformed->insert(array_name);
    }
}

void ArrayFlattener::transformParameters(SgFunctionParameterList *parameters) {
    SgInitializedNamePtrList &args = parameters->get_args();

    for (auto &it : args) {
        SgInitializedName *arg = isSgInitializedName(it);

        if (AstUtil::isImageType(arg->get_type()) || AstUtil::isBlockType(arg->get_type())) {
            SgType *baseType;
            switch (kernelInfo.getPixelType(arg->get_name().str())) {
                case FLOAT:
                    baseType = buildFloatType();
                    break;
                case INT:
                    baseType = buildIntType();
                    break;
                case UCHAR:
                    baseType = buildUnsignedCharType();
                    break;
                case DOUBLE:
                    baseType = buildDoubleType();
                    break;
                default:
                    baseType = buildFloatType();
                    break;
            }

            arg->set_type(buildPointerType(baseType));
        }
    }
}
