// Copyright (c) 2016, Thomas L. Falch
// For conditions of distribution and use, see the accompanying LICENSE and README files

// This file is part of the ImageCL source-to-source compiler
// developed at the Norwegian University of Science and technology


#include "astutil.h"

#include <utility>

using namespace std;
using namespace SageInterface;

void AstUtil::fixUniqueNameAttributes(SgProject *project) {

    Rose_STL_Container<SgNode *> expressionNodes = NodeQuery::querySubTree(project, V_SgExpression);

    for (SgNode *expressionNode : expressionNodes) {

        SgExpression *expression = isSgExpression(expressionNode);
        AstAttributeMechanism *attributeMechanism = expression->get_attributeMechanism();
        if (attributeMechanism != NULL) {
            expression->removeAttribute("UniqueNameAttribute");

        }
    }
    annotateExpressionsWithUniqueNames(project);
}

bool AstUtil::isConstantMemCandidate(SgType *type) {
    int size = 1;
    SgType *typeTemp = NULL;
    SgArrayType *arrayTemp = isSgArrayType(type);
    if (!arrayTemp) {
        return false;
    }
    while (arrayTemp) {

        SgIntVal *intIndex = isSgIntVal(arrayTemp->get_index());
        SgCastExp *cast = isSgCastExp(arrayTemp->get_index());
        if (cast) {
            intIndex = isSgIntVal(cast->get_operand());
        }
        if (intIndex) {
            size *= intIndex->get_value();
        } else {
            return false;
        }
        typeTemp = arrayTemp->get_base_type();
        arrayTemp = isSgArrayType(typeTemp);
    }
    return size < 256;
}


bool AstUtil::isImageType(SgType *type) {
    return type->unparseToString().find("**") != string::npos && type->unparseToString().find("***") == string::npos;
}

bool AstUtil::isBlockType(SgType *type) {
    return type->unparseToString().find("***") != string::npos;
}

SgFunctionDeclaration *AstUtil::getFunctionDeclaration(SgProject *project, string functionName) {
    Rose_STL_Container<SgNode *> funcDefNodes = NodeQuery::querySubTree(project, V_SgFunctionDeclaration);

    for (SgNode *funcDefNode : funcDefNodes) {

        SgFunctionDeclaration *funcDef = isSgFunctionDeclaration(funcDefNode);

        if (funcDef) {
            if (funcDef->get_name() == functionName) {
                return funcDef;
            }
        }
    }
    exit(58);
}

bool AstUtil::is2DArrayRoot(SgNode *node) {
    SgPntrArrRefExp *arrRef = isSgPntrArrRefExp(node);
    if (!arrRef) {
        return false;
    }

    SgNode *parent = arrRef->get_parent();
    if (isSgVarRefExp(parent)) {
        return false;
    }

    SgVarRefExp *array = nullptr;
    SgPntrArrRefExp *temp = arrRef;

    int dimensions = 0;
    while (array == nullptr && temp != nullptr) {
        array = isSgVarRefExp(temp->get_lhs_operand());
        temp = isSgPntrArrRefExp(temp->get_lhs_operand());
        dimensions++;
    }

    return dimensions == 2;
}

bool AstUtil::is3DArrayRoot(SgNode *node) {
    SgPntrArrRefExp *arrRef = isSgPntrArrRefExp(node);
    if (!arrRef) {
        return false;
    }

    SgNode *parent = arrRef->get_parent();
    if (isSgVarRefExp(parent)) {
        return false;
    }

    SgVarRefExp *array = nullptr;
    SgPntrArrRefExp *temp = arrRef;

    int dimensions = 0;
    while (array == nullptr && temp != nullptr) {
        array = isSgVarRefExp(temp->get_lhs_operand());
        temp = isSgPntrArrRefExp(temp->get_lhs_operand());
        dimensions++;
    }

    return dimensions == 3;
}

string AstUtil::getArrayName(SgPntrArrRefExp *arrRef) {
    SgVarRefExp *array = nullptr;
    SgPntrArrRefExp *temp = arrRef;

    int dimensions = 0;
    while (array == nullptr && temp != nullptr) {
        array = isSgVarRefExp(temp->get_lhs_operand());
        temp = isSgPntrArrRefExp(temp->get_lhs_operand());
        dimensions++;
    }

    return array->get_symbol()->get_name().str();
}


SgStatement *AstUtil::getParentStatement(SgNode *node) {
    SgNode *parentStatementNode = node->get_parent();
    while (!isSgStatement(parentStatementNode)) {
        parentStatementNode = parentStatementNode->get_parent();
    }
    SgStatement *parentStatement = isSgStatement(parentStatementNode);
    return parentStatement;
}


set<string> *AstUtil::getArrayNamesReferenced(SgNode *node) {
    auto *arrayNames = new set<string>();

    Rose_STL_Container<SgNode *> arrRefNodes = NodeQuery::querySubTree(node, V_SgPntrArrRefExp);

    for (SgNode *arrRefNode: arrRefNodes) {

        SgNode *parent = arrRefNode->get_parent();
        if (isSgPntrArrRefExp(parent)) {
            continue;
        }

        SgPntrArrRefExp *arrRef = isSgPntrArrRefExp(arrRefNode);

        string arrayName = AstUtil::getArrayName(arrRef);

        arrayNames->insert(arrayName);
    }

    return arrayNames;
}

set<string> *AstUtil::findArrayArguments(SgProject *project, string kernelName) {
    auto *arrayArguments = new set<string>();

    SgFunctionDeclaration *funcDef = AstUtil::getFunctionDeclaration(project, std::move(kernelName));
    const vector<SgInitializedName *> &kernelArguments = funcDef->get_parameterList()->get_args();

    for (const SgInitializedName *kernelArgument : kernelArguments) {

        auto isPointer = static_cast<bool>(isSgPointerType(kernelArgument->get_type()));
        auto isArray = static_cast<bool>(isSgArrayType(kernelArgument->get_type()));
        auto isImage = AstUtil::isImageType(kernelArgument->get_type());
        auto isBlock = AstUtil::isBlockType(kernelArgument->get_type());
        if ((isPointer || isArray || isImage || isBlock)) {
            arrayArguments->insert(kernelArgument->get_name().str());
        }
    }

    return arrayArguments;
}
