// Copyright (c) 2016, Thomas L. Falch
// For conditions of distribution and use, see the accompanying LICENSE and README files

// This file is part of the ImageCL source-to-source compiler
// developed at the Norwegian University of Science and technology


#include "rose.h"

#include "astutil.h"
#include "pragma.h"
#include "type.h"
#include "filehandler.h"
#include "kernelinfo.h"
#include "argument.h"

using namespace SageInterface;
using namespace std;


KernelInfo::KernelInfo(SgProject *project, Settings settings) : project(project), settings(settings) {
    this->constantArrays = NULL;
    this->readOnlyArrays = NULL;
    this->imageArrays = NULL;
    this->pragmas = NULL;
    this->gridArrays = NULL;
    this->pixelTypes = NULL;
    this->writeOnlyArrays = NULL;
    this->allArrays = NULL;
}

void KernelInfo::findKernelName() {
    Rose_STL_Container<SgNode *> nodes = NodeQuery::querySubTree(project, V_SgFunctionDeclaration);
    for (SgNode *node : nodes) {
        SgFunctionDeclaration *funcDef = isSgFunctionDeclaration(node);

        if (!funcDef->isCompilerGenerated())
            kernelName = funcDef->get_name().getString();
    }
}


void KernelInfo::findImageAndBlockArrays() {
    this->imageArrays = new set<string>();
    this->pixelTypes = new map<string, BaseType>();
    SgFunctionDeclaration *funcDec = AstUtil::getFunctionDeclaration(project, kernelName);

    const vector<SgInitializedName *> &arguments = funcDec->get_args();

    // Find image or block types in arguments.
    for (const SgInitializedName *argument : arguments) {
        if (AstUtil::isImageType(argument->get_type())) {
            imageArrays->insert(argument->get_name().str());

            pixelTypes->insert(make_pair(argument->get_name().str(),
                                         Argument::convertType(argument->get_type(), settings).baseType));
        } else if (AstUtil::isBlockType(argument->get_type())) {
            blockArrays.insert(argument->get_name().str());

            pixelTypes->insert(make_pair(argument->get_name().str(),
                                         Argument::convertType(argument->get_type(), settings).baseType));
        }
    }

    // Find image or block types in the code.
    // We need to find all images or blocks defined in the code and look at its expression.
    // Only images or blocks defined by another image or block is valid, i.e. `Block<T> x = data + offset` where data is a `Block<T>`.
    // Then we need to map `x` to `data` with an offset so all optimizations are actually done on `data` and not on `x`.
    for (auto n : NodeQuery::querySubTree(project, V_SgVariableDeclaration)) { // V_SgVariableDeclaration + V_SgAssignInitializer
        auto *varDecl = isSgVariableDeclaration(n);

        for (auto &a : varDecl->get_variables()) {
            if (AstUtil::isImageType(a->get_type())) {
                for (auto m : NodeQuery::querySubTree(a->get_parent(), V_SgAddOp)) {
                    auto *add = isSgAddOp(m);
                    auto *lhs = add->get_lhs_operand();
                    auto *rhs = add->get_rhs_operand();
                    if (isImageArray(lhs->unparseToString())) {
                        aliasArrays[a->unparseToString()] =  std::make_pair(lhs->unparseToString(), rhs);
                    } else if (isImageArray(rhs->unparseToString())) {
                        aliasArrays[a->unparseToString()] = std::make_pair(rhs->unparseToString(), lhs);
                    }
                    pixelTypes->insert(make_pair(a->get_name().str(),
                                                 Argument::convertType(a->get_type(), settings).baseType));
                }
            } else if (AstUtil::isBlockType(a->get_type())) {
                for (auto m : NodeQuery::querySubTree(a->get_parent(), V_SgAddOp)) {
                    auto *add = isSgAddOp(m);
                    auto *lhs = add->get_lhs_operand();
                    auto *rhs = add->get_rhs_operand();
                    if (isBlockArray(lhs->unparseToString())) {
                        aliasArrays[a->unparseToString()] =  std::make_pair(lhs->unparseToString(), rhs);
                    } else if (isBlockArray(rhs->unparseToString())) {
                        aliasArrays[a->unparseToString()] = std::make_pair(rhs->unparseToString(), lhs);
                    }
                    pixelTypes->insert(make_pair(a->get_name().str(),
                                                 Argument::convertType(a->get_type(), settings).baseType));
                }
            }
        }
    }
}

set<string> *KernelInfo::findArraysReadFromWrittenTo(bool findArraysReadFrom, bool findArraysWrittenTo) {
    auto *arrays = new set<string>();
    if (!findArraysReadFrom && !findArraysWrittenTo) {
        return arrays;
    }

    for (auto *node : NodeQuery::querySubTree(project, V_SgPntrArrRefExp)) {
        SgPntrArrRefExp *arrayRef = isSgPntrArrRefExp(node);

        if (isSgPntrArrRefExp(arrayRef->get_parent())) {
            continue;
        }

        string arrayName = AstUtil::getArrayName(arrayRef);
        string aliasName = arrayName;
        // If arrayRef is an alias of another object, add the original object.
        for (const auto &alias : this->aliasArrays) {
            if (arrayName == alias.first) {
                aliasName = alias.second.first;
                break;
            }
        }

        if (arrayRef->get_lvalue()) {
            if (findArraysWrittenTo) {
                arrays->insert(arrayName);
                arrays->insert(aliasName);
            }
        } else {
            if (findArraysReadFrom) {
                arrays->insert(arrayName);
                arrays->insert(aliasName);
            }
        }
    }

    return arrays;
}

void KernelInfo::findReadOnlyArrays() {
    readOnlyArrays = new set<string>();

    set<string> *arraysWrittenTo = findArraysReadFromWrittenTo(false, true);

    std::set_difference(allArrays->begin(), allArrays->end(), arraysWrittenTo->begin(), arraysWrittenTo->end(),
                        inserter(*readOnlyArrays, readOnlyArrays->begin()));
}

void KernelInfo::findWriteOnlyArrays() {
    writeOnlyArrays = new set<string>();

    set<string> *arraysReadFrom = findArraysReadFromWrittenTo(true, false);

    std::set_difference(allArrays->begin(), allArrays->end(), arraysReadFrom->begin(), arraysReadFrom->end(),
                        inserter(*writeOnlyArrays, writeOnlyArrays->begin()));
}


void KernelInfo::parsePragmas() {
    this->pragmas = new set<Pragma>();
    this->gridArrays = new set<string>();
    Rose_STL_Container<SgNode *> pragmas = NodeQuery::querySubTree(project, V_SgPragmaDeclaration);

    for (SgNode *n: pragmas) {
        SgPragmaDeclaration *pragmaDecl = isSgPragmaDeclaration(n);
        SgPragma *pragma = pragmaDecl->get_pragma();
        string pragmaString = pragma->get_pragma();

        const auto location = pragmaString.find("clite");
        if (location == string::npos || location < pragmaString.size()) {

            try {
                Pragma p(pragmaString);
                this->pragmas->insert(p);
            }
            catch (...) {
                cerr << "Error parsing pragma. Exiting..." << endl;
                exit(-1);
            }
        }
    }
}

void KernelInfo::findConstantArrays() {
    constantArrays = new set<string>();
    SgFunctionDeclaration *funcDec = AstUtil::getFunctionDeclaration(project, kernelName);
    vector<SgInitializedName *> arguments = funcDec->get_args();

    for (SgInitializedName *argument : arguments) {
        bool isCandidate = AstUtil::isConstantMemCandidate(argument->get_type());
        bool isReadOnly = this->readOnlyArrays->count(argument->get_name().str()) == 1;
        bool isConstCandPragma = false;
        for (Pragma p : *(pragmas)) {
            if (p.option == CONSTANT_MEM_CAND)
                if (p.values.count(argument->get_name().str()) == 1)
                    isConstCandPragma = true;
        }

        if (isReadOnly && (isCandidate || isConstCandPragma)) {
            constantArrays->insert(argument->get_name().str());
        }
    }
}


void KernelInfo::findUnrollableLoops() {
    forLoops = new vector<pair<int, vector<int> *>>();
    Rose_STL_Container<SgNode *> forLoopNodes = NodeQuery::querySubTree(project, V_SgForStatement);

    for (SgNode *forLoopNode : forLoopNodes) {
        SgForStatement *forLoop = isSgForStatement(forLoopNode);

        SgInitializedName *ivar = NULL;
        SgExpression *lb = NULL;
        SgExpression *ub = NULL;
        SgExpression *step = NULL;
        bool isIncremental, isInclusiveUpperBound;
        SgStatement *lbody = NULL;
        bool isCannonical = isCanonicalForLoop(forLoop, &ivar, &lb, &ub, &step, &lbody, &isIncremental,
                                               &isInclusiveUpperBound);

        SgIntVal *lb_int = isSgIntVal(lb);
        SgMinusOp *lbMinus = isSgMinusOp(lb);
        if (lbMinus) {
            lb_int = isSgIntVal(lbMinus->get_operand());
        }
        SgIntVal *ub_int = isSgIntVal(ub);
        SgIntVal *step_int = isSgIntVal(step);
        if (isCannonical && lb_int && ub_int && step_int) {
            pair<int, vector<int> *> forLoopEntry = make_pair(forLoop->get_file_info()->get_line(), new vector<int>());
            int upper = ub_int->get_value();
            int lower = lb_int->get_value();
            if (lbMinus) {
                lower *= -1;
            }
            int nIterations = (upper - lower);
            if (isInclusiveUpperBound) {
                nIterations++;
            }
            int s = step_int->get_value();
            nIterations /= s;

            for (int i = 2; i <= nIterations / 2; i++) {
                if (nIterations % i == 0) {
                    forLoopEntry.second->push_back(i);
                }
            }
            forLoopEntry.second->push_back(nIterations);
            forLoops->push_back(forLoopEntry);
        }
    }
}

void KernelInfo::setGridFromPragmas() {
    bool foundGrid = false;
    bool foundGridSize = false;

    for (Pragma p : *pragmas) {
        if (p.option == GRID) {
            foundGrid = true;

            for (const string &gridArray : p.values) {
                if (!isImageArray(gridArray) && !isBlockArray(gridArray)) {
                    cerr << "ERROR: Grid array must be Image or Block (" << gridArray << ", " << isImageArray(gridArray) << ", " << isBlockArray(gridArray) << "). Exiting..." << endl;
                    exit(-1);
                }
                this->gridArrays->insert(gridArray);
            }
        }
    }

    for (Pragma p : *pragmas) {
        if (p.option == GRID_SIZE) {
            foundGridSize = true;
            this->constGridSize = true;
            this->gridSizeX = p.x;
            this->gridSizeY = p.y;
            this->gridSizeZ = p.z;
        }
    }

    if (!(foundGrid || foundGridSize)) {
        cerr << "ERROR: Grid size or array must be specified. Exiting..." << endl;
        exit(-1);
    }

    if (foundGrid && foundGridSize) {
        cerr << "WARNING: Both grid size and grid arrays set. Ignoring grid size." << endl;
        this->constGridSize = false;
    }
}

void KernelInfo::setBoundaryConditions() {
    this->boundaryConditions = new map<string, BoundaryCondition>();
    for (string s : *imageArrays) {
        boundaryConditions->insert(make_pair(s, settings.boundaryCondition));
    }
    for (string s : blockArrays) {
        boundaryConditions->insert(make_pair(s, settings.boundaryCondition));
    }

    for (Pragma p : *pragmas) {
        if (p.option == BOUNDARY_COND) {
            for (pair<string, string> ps : p.pairValues) {
                if (ps.second.compare("constant") == 0) {
                    (*boundaryConditions)[ps.first] = BoundaryCondition::CONSTANT;
                }
                if (ps.second.compare("clamped") == 0) {
                    (*boundaryConditions)[ps.first] = BoundaryCondition::CLAMPED;
                }
                if (ps.second.compare("none") == 0) {
                    (*boundaryConditions)[ps.first] = BoundaryCondition::NONE;
                }
            }
        }
    }
}


BaseType KernelInfo::getPixelType(string arr) {
    if (pixelTypes->count(arr) == 1) {
        return pixelTypes->at(arr);
    } else {
        return settings.defaultPixelType;
    }
}

bool KernelInfo::needsMpiBroadcast(string argumentName) {
    if (allArrays->count(argumentName) == 1) {
        return !needsMpiScatter(argumentName);
    }
    return false;
}

bool KernelInfo::needsMpiScatter(string argumentName) {

    //TODO: REMOVE this requirement, it is to strict
    if (gridArrays->count(argumentName) == 0) {
        return false;
    }

    if (footprintTable.count(argumentName) == 0) {
        return false;
    }
    Footprint fp = footprintTable.at(argumentName);

    return !fp.threadStatic;

}

bool KernelInfo::needsGridPosArg() {
    for (string s : *allArrays) {
        if (needsMpiScatter(s) && getBoundaryConditionForArray(s) == BoundaryCondition::CLAMPED)
            return true;
        if (settings.generateOMP) {
            if (needsMpiScatter(s) && getBoundaryConditionForArray(s) == BoundaryCondition::CONSTANT) {
                return true;
            }
        }
    }
    return false;
}

HaloSize KernelInfo::getHaloSize(string array) {
    return footprintTable.at(array).computeHaloSize();
}


void KernelInfo::scanForInfo() {
    parsePragmas();

    findKernelName();

    findUnrollableLoops();

    this->allArrays = AstUtil::findArrayArguments(project, kernelName);
    for (auto node : NodeQuery::querySubTree(project, V_SgVariableDeclaration)) {
        auto *varDecl = isSgVariableDeclaration(node);
        for (auto &a : varDecl->get_variables()) {
            if (AstUtil::isImageType(a->get_type()) || AstUtil::isBlockType(a->get_type())) {
                this->allArrays->insert(a->unparseToString());
            }
        }
    }
    findImageAndBlockArrays();
    findReadOnlyArrays();
    findWriteOnlyArrays();
    findConstantArrays();

    setGridFromPragmas();
    setBoundaryConditions();

    FootprintFinder fpf;
    footprintTable = fpf.findFootprints(project, this);
}

string KernelInfo::getAGridArray() {
    return *gridArrays->begin();
}

bool KernelInfo::isReadOnlyArray(string arrayName) {
    return this->readOnlyArrays->count(arrayName) == 1;
}

bool KernelInfo::isWriteOnlyArray(string arrayName) {
    return this->writeOnlyArrays->count(arrayName) == 1;
}

BoundaryCondition KernelInfo::getBoundaryConditionForArray(string array) {
    return this->boundaryConditions->at(array);
}

bool KernelInfo::isGridArray(string arrayName) {
    return this->gridArrays->count(arrayName) == 1;
}

bool KernelInfo::isImageArray(string arrayName) {
    return this->imageArrays->count(arrayName) == 1;
}

string KernelInfo::getKernelName() {
    return kernelName;
}

set<string> *KernelInfo::getGridArrays() {
    return gridArrays;
}

map<string, Footprint> KernelInfo::getFootprintTable() {
    return footprintTable;
}

set<string> *KernelInfo::getReadOnlyArrays() {
    return readOnlyArrays;
}

set<string> *KernelInfo::getConstantArrays() {
    return constantArrays;
}

set<string> *KernelInfo::getImageArrays() {
    return imageArrays;
}

set<Pragma> *KernelInfo::getPragmas() {
    return this->pragmas;
}


void KernelInfo::printKernelInfo() {
    char esc_char = 27;
    cout << endl;
    cout << esc_char << "[1m" << "== Kernel Info ==" << esc_char << "[0m" << endl;
    cout << "Kernel name: " << this->kernelName << endl;
    cout << "Footprint table:" << endl;
    for (pair<string, Footprint> f : this->footprintTable) {
        cout << "  " << f.first << " : " << f.second.str() << endl;
    }

    cout << "All arrays: ";
    for (const auto &s : *allArrays) {
        cout << s << " ";
    }
    cout << endl;

    cout << "Read only arrays: ";
    for (string s : *readOnlyArrays) {
        cout << s << " ";
    }
    cout << endl;

    cout << "Write only arrays: ";
    for (string s : *writeOnlyArrays) {
        cout << s << " ";
    }
    cout << endl;

    cout << "Constant arrays: ";
    for (string s : *constantArrays) {
        cout << s << " ";
    }
    cout << endl;

    cout << "Image arrays: ";
    for (string s : *imageArrays) {
        cout << s << " ";
    }
    cout << endl;

    cout << "Block arrays: ";
    for (string s: blockArrays) {
        cout << s << " ";
    }
    cout << endl;

    cout << "Alias arrays: ";
    for (std::pair<string, std::pair<string, SgExpression*>> a: aliasArrays) {
        cout << a.first << ": " << a.second.first << " + " << a.second.second->unparseToString() << ", ";
    }
    cout << endl;

    cout << "Grid arrays: ";
    for (string s : *gridArrays) {
        cout << s << " ";
    }
    cout << endl;


    cout << "For loops: " << endl;
    for (pair<int, vector<int> *> forLoopEntry : *forLoops) {
        cout << forLoopEntry.first - FileHandler::preambleLength << ":";
        for (int i : *(forLoopEntry.second)) {
            cout << i << ",";
        }
        cout << endl;
    }
    cout << "Pragmas:" << endl;

    for (Pragma p: *pragmas) {
        cout << "  " << p.str() << endl;
    }

    cout << "Boundary conditions:" << endl;
    for (pair<string, BoundaryCondition> psb : *boundaryConditions) {
        cout << "  " << psb.first << ": ";
        if (psb.second == BoundaryCondition::CLAMPED)
            cout << "CLAMPED";
        if (psb.second == BoundaryCondition::CONSTANT)
            cout << "CONSTANT";
        if (psb.second == BoundaryCondition::NONE)
            cout << "NONE";
        cout << endl;
    }

    cout << "Pixel types:" << endl;
    for (pair<string, BaseType> psb : *pixelTypes) {
        cout << "  " << psb.first << ": " << Type::baseTypeToString(psb.second) << endl;
    }

    cout << endl;
}

vector<pair<int, vector<int> *>> *KernelInfo::getForLoops() {
    return this->forLoops;
}
