// Copyright (c) 2016, Thomas L. Falch
// For conditions of distribution and use, see the accompanying LICENSE and README files

// This file is part of the ImageCL source-to-source compiler
// developed at the Norwegian University of Science and technology


#include <iostream>
#include <fstream>
#include <vector>
#include <set>

#include "rose.h"

#include "genericDataflowCommon.h"

#include "globalremover.h"
#include "naivecoarsener.h"
#include "imagememtransformer.h"
#include "arrayflattener.h"
#include "kernelfinder.h"
#include "wrappergenerator.h"
#include "localmemtransformer.h"
#include "boundryguardinserter.h"
#include "constantmemtransformer.h"
#include "loopunroller.h"
#include "fastwrappergenerator.h"
#include "filehandler.h"
#include "indexchanger.h"
#include "register_caching.h"

using namespace std;
using namespace SageBuilder;
using namespace SageInterface;
using namespace CommandlineProcessing;


void patch_ast(SgProject *project, const Settings &settings, KernelInfo &kernelInfo, const Parameters &params, vector<Argument> *arguments) {
    auto varRefs = NodeQuery::querySubTree(project, V_SgVarRefExp);
    //This has to be the ugliest hack ever, somehow needed to fix AST due to template types for image arrays
    //Should be moved to a separate function or something
    for (SgNode *node : varRefs) {
        SgVarRefExp *varRef = isSgVarRefExp(node);

        if (varRef->get_symbol()->get_name() == "idx") {
            //cout << varRef->get_symbol()->get_name() << endl;

            SgScopeStatement *scope = getEnclosingScope(varRef);
            SgVarRefExp *newVarRef = buildVarRefExp("idx", scope);

            replaceExpression(varRef, newVarRef);
        } else if (varRef->get_symbol()->get_name() == "idy") {
            //cout << varRef->get_symbol()->get_name() << endl;

            SgScopeStatement *scope = getEnclosingScope(varRef);
            SgVarRefExp *newVarRef = buildVarRefExp("idy", scope);

            replaceExpression(varRef, newVarRef);
        } else if (varRef->get_symbol()->get_name() == "idz") {
            SgScopeStatement *scope = getEnclosingScope(varRef);
            SgVarRefExp *newVarRef = buildVarRefExp("idz", scope);

            replaceExpression(varRef, newVarRef);
        }
    }

    auto tempDecl = NodeQuery::querySubTree(project, V_SgTemplateTypedefDeclaration);

    for (SgNode *node : tempDecl) {
        SgTemplateTypedefDeclaration *temptypdef = isSgTemplateTypedefDeclaration(node);
        removeStatement(temptypdef);
    }

    // Instead of sending in GS_X and GS_Y as options like in OpenCL, we send them in as input.
    // Also remove parameter if it is texture memory.
    // Also declare texture memory.
    // :(
    if (settings.language == Language::Cuda) {
        auto nodes = NodeQuery::querySubTree(project, V_SgFunctionDeclaration);
        for (SgNode *node : nodes) {
            SgFunctionDeclaration *funcDef = isSgFunctionDeclaration(node);
            if (funcDef->get_name().str() == kernelInfo.getKernelName()) {
                if (auto *parent = isSgGlobal(funcDef->get_parent())) {
                    for (auto &arg : *arguments) {
                        if (params.imageMemArrays->count(arg.name) == 1 && !kernelInfo.isBlockArray(arg.name)) {
                            const auto t = string("texture<") + Type::baseTypeToString(kernelInfo.getPixelType(arg.name)) + ", 2>";
                            auto *texType = buildOpaqueType(t, parent->get_scope());
                            auto *texDecl = buildVariableDeclaration(arg.name + "_tex", texType, nullptr, parent->get_scope());
                            parent->prepend_declaration(texDecl);
                        } else if (params.constantMemArrays->count(arg.name) == 1) {
                            // Size of constant memory in cuda must be set statically, and we dont know the size at this moment.
                            // NOTE: Max const memory size is ~64kB so we could do 64000 / sizeof(type)
                            // Current solution: set it to 0 and let user manually set the correct size.
                            auto *constType = buildOpaqueType(string("__constant__ ") + Type::baseTypeToString(kernelInfo.getPixelType(arg.name)), parent->get_scope());

                            if (kernelInfo.isConstGridSize()) {
                                auto *constDecl = buildVariableDeclaration(
                                        arg.name + "[" + std::to_string(kernelInfo.gridSizeX * kernelInfo.gridSizeY * kernelInfo.gridSizeZ) + "]",
                                        constType, nullptr, parent->get_scope());
                                parent->prepend_declaration(constDecl);
                            } else {
                                auto *constDecl = buildVariableDeclaration(arg.name + "[0]", constType, nullptr, parent->get_scope());
                                parent->prepend_declaration(constDecl);
                            }

                            //arg.name = arg.name + "_host";

                        }
                        if (params.imageMemArrays->count(arg.name) == 1 && kernelInfo.isBlockArray(arg.name)) {
                            /// A function that allows reading double from texture memory.
                            //newFile << "__device__ double tex3D_double(cudaTextureObject_t texObj, float x, float y, float z) {\n";
                            //newFile << "\tint2 a = tex3D<int2>(texObj, x, y, z);\n";
                            //newFile << "\treturn __hiloint2double(a.y, a.x);\n";
                            //newFile << "}\n";

                            auto *texObj = buildInitializedName("texObj", buildOpaqueType("cudaTextureObject_t", parent->get_scope()));
                            auto *x = buildInitializedName("x", buildFloatType());
                            auto *y = buildInitializedName("y", buildFloatType());
                            auto *z = buildInitializedName("z", buildFloatType());
                            auto *fn = buildDefiningFunctionDeclaration("tex3D_double", buildDoubleType(), buildFunctionParameterList(texObj, x, y, z), parent->get_scope());
                            fn->get_functionModifier().setCudaDevice();
                            auto *body = fn->get_definition()->get_body();
                            auto *scope = fn->get_scope();

                            auto *tex3d = buildFunctionCallExp("tex3D<int2>",
                                                 buildOpaqueType("int2", scope),
                                                 buildExprListExp(buildVarRefExp("texObj", scope), buildVarRefExp("x", scope), buildVarRefExp("y", scope), buildVarRefExp("z", scope)),
                                                 scope);
                            auto* a = buildVariableDeclaration("a", buildOpaqueType("int2", scope), buildAssignInitializer(tex3d, buildOpaqueType("int2", scope)), scope);
                            auto *cast = buildFunctionCallExp("__hiloint2double",
                                                 buildDoubleType(),
                                                 buildExprListExp(buildOpaqueVarRefExp("a.y", scope), buildOpaqueVarRefExp("a.x", scope)),
                                                 scope);
                            auto *ret = buildReturnStmt(cast);
                            body->append_statement(a);
                            body->append_statement(ret);
                            parent->prepend_declaration(fn);
                        }
                    }
                }
                auto *param_list = funcDef->get_parameterList();
                auto &args = param_list->get_args();
                args.erase(remove_if(args.begin(), args.end(), [&](SgInitializedName *arg) {
                    return (params.imageMemArrays->count(arg->get_name()) && !kernelInfo.isBlockArray(arg->get_name()))
                           || params.constantMemArrays->count(arg->get_name());
                }), args.end());
                param_list->append_arg(buildInitializedName("GS_X", buildIntType()));
                param_list->append_arg(buildInitializedName("GS_Y", buildIntType()));
                param_list->append_arg(buildInitializedName("GS_Z", buildIntType()));
            }
        }

        SgFunctionDeclaration *funcDef = AstUtil::getFunctionDeclaration(project, kernelInfo.getKernelName());
        for (SgInitializedName *arg : funcDef->get_args()) {
            if (params.imageMemArrays->count(arg->get_name().str()) == 1 && kernelInfo.isBlockArray(arg->get_name().str())) {
                auto *tex = buildOpaqueType("cudaTextureObject_t", funcDef->get_definition()->get_body());
                arg->set_type(tex);
            }
        }
    }
}

void generateCCode(SgProject *project, const KernelInfo &kernelInfo, const Settings &settings, Parameters params) {
    GlobalVariableRemover globalVarRemover;
    globalVarRemover.traverseInputFiles(project, preorder);

    ArrayFlattener arrayFlattener(project, params, kernelInfo, settings);
    arrayFlattener.transform();

    ArgumentHandler argumentHandler(project, kernelInfo, params, settings);
    argumentHandler.addCArguments();

    auto tempDecl = NodeQuery::querySubTree(project, V_SgTemplateTypedefDeclaration);

    for (SgNode *node : tempDecl) {
        SgTemplateTypedefDeclaration *temptypdef = isSgTemplateTypedefDeclaration(node);
        removeStatement(temptypdef);
    }

    project->unparse();
}

int main(int argc, char **argv) {

    ROSE_INITIALIZE;
    auto commandLineArgs = generateArgListFromArgcArgv(argc, argv);
    auto fileNames = generateSourceFilenames(commandLineArgs, false);
    if (fileNames.size() > 1) {
        cout << "ERROR: can only handle one input file. Exiting..." << endl;
        exit(-1);
    }

    const bool generateParamSpec = isOption(commandLineArgs, "-clite:", "g", false);
    const bool generateC = isOption(commandLineArgs, "-clite:", "c", false);
    const bool generateCuda = isOption(commandLineArgs, "-clite:", "cuda", false);
    const bool generateCL = isOption(commandLineArgs, "-clite:", "opencl", false);
    const bool verbose = isOption(commandLineArgs, "-clite:", "v", false);

    Settings settings;
    settings.readSettingsFromFile("settings.txt");
    settings.inputBaseName = FileHandler::getFileNameBase(fileNames[0]);
    if (generateC) {
        settings.language = Language::C;
    } else if (generateCuda) {
        settings.language = Language::Cuda;
    } else if (generateCL) {
        settings.language = Language::OpenCL;
    }

    if (verbose) {
        settings.printSettings();
    }

    FileHandler::fixFiles(fileNames, settings);
    char **newArgv = FileHandler::fixFileNames(argc, argv, fileNames);

    SgProject *project = frontend(argc, newArgv);

    KernelInfo kernelInfo(project, settings);
    kernelInfo.scanForInfo();
    if (verbose) {
        kernelInfo.printKernelInfo();
    }

    Parameters params;
    if (generateParamSpec) {
        params.generateParameterSpecification(kernelInfo);
        generateDOT(*project);
        return 0;
    }

    params.setDefaultParameters();
    if (settings.language != Language::C) {
        params.readParametersFromFile(kernelInfo, "config.txt");
        params.setParametersFromPragmas(kernelInfo.getPragmas());
    }
    params.validateParameters(kernelInfo);
    if (verbose) {
        params.printParameters();
    }

    if (settings.language == Language::C) {
        generateCCode(project, kernelInfo, settings, params);
        FileHandler::cleanUpFiles(fileNames, settings, false);
    } else {
        if (verbose) {
            cout << "unrolling loops" << endl;
        }
        LoopUnroller loopUnroller(project, params, kernelInfo);
        loopUnroller.traverseInputFiles(project, postorder);
        loopUnroller.unrollLoops();

        if (verbose) {
            cout << "removing global variables" << endl;
        }

        GlobalVariableRemover globalVarRemover;
        globalVarRemover.traverseInputFiles(project, preorder);

        if (verbose) {
            cout << "naive coarsener" << endl;
        }

        NaiveCoarsener naiveCoarsener(params, kernelInfo, settings);
        naiveCoarsener.traverseInputFiles(project, preorder);
        fixVariableReferences(project);

        if (params.useLocalMem()) {
            if (verbose) {
                cout << "applying local memory" << endl;
            }
            LocalMemTransformer localMemTransformer(project, params, kernelInfo, settings,
                                                    naiveCoarsener.getOriginalFunctionBody());
            localMemTransformer.transform();
        }

        if (params.useRegCache() && settings.language == Language::Cuda) {
            if (verbose) {
                cout << "applying register caching" << endl;
            }
            auto reg = CudaRegisterCaching(project, kernelInfo, params, settings,
                                           naiveCoarsener.getOriginalFunctionBody());
            reg.transform();
        }

        IndexChanger indexChanger(kernelInfo, params, settings);
        if (settings.generateMPI || settings.generateOMP) {
            if (verbose) {
                cout << "index changer" << endl;
            }
            indexChanger.replaceIndices(project);
        }

        if (verbose) {
            cout << "inserting boundary guard" << endl;
        }

        BoundryGuardInserter boundryGuardInserter(kernelInfo, params, project, settings);
        boundryGuardInserter.insertBoundryGuards();

        if (settings.generateMPI || settings.generateOMP) {
            indexChanger.addPaddings(project);
        }

        if (params.useImageMem()) {
            if (verbose) {
                cout << "applying image memory" << endl;
            }
            ImageMemTransformer imageMemTransformer(project, params, kernelInfo, settings);
            imageMemTransformer.transform();
        }

        if (verbose) {
            cout << "flattening array" << endl;
        }

        ArrayFlattener arrayFlattener(project, params, kernelInfo, settings);
        arrayFlattener.transform();

        if (params.useConstantMem()) {
            if (verbose) {
                cout << "applying constant memory" << endl;
            }

            ConstantMemTransformer constantMemTransformer(project, kernelInfo, params);
            constantMemTransformer.transform(settings);
        }

        if (verbose) {
            cout << "kernel finder" << endl;
        }

        KernelFinder kernelFinder(project, params, kernelInfo);
        kernelFinder.transform(settings.language);

        if (verbose) {
            cout << "argument handler" << endl;
        }

        ArgumentHandler argumentHandler(project, kernelInfo, params, settings);
        vector<Argument> *arguments = argumentHandler.addAndGetArguments();

        if (verbose) {
            cout << "patching ast" << endl;
        }

        patch_ast(project, settings, kernelInfo, params, arguments);

        if (settings.generateStandalone) {
            if (verbose) {
                cout << "generating standalone" << endl;
            }
            WrapperGenerator wrapperGenerator(settings.inputBaseName + "_wrapper." + (settings.language == Language::OpenCL ? "c" : "cu"), arguments, params, kernelInfo, settings);
            wrapperGenerator.generate();
        }
        if (settings.generateFAST) {
            if (verbose) {
                cout << "generating FAST" << endl;
            }
            FASTWrapperGenerator fastWrapperGenerator("FastWrapper", arguments, params, kernelInfo);
            fastWrapperGenerator.generate();
        }

        generateDOT(*project);

        project->unparse();

        FileHandler::cleanUpFiles(fileNames, settings, true);
    }
}
