// Copyright (c) 2016, Thomas L. Falch
// For conditions of distribution and use, see the accompanying LICENSE and README files

// This file is part of the ImageCL source-to-source compiler
// developed at the Norwegian University of Science and technology


#include "naivecoarsener.h"

using namespace std;
using namespace SageBuilder;
using namespace SageInterface;

NaiveCoarsener::NaiveCoarsener(Parameters params, KernelInfo kernelInfo, Settings settings) : params(params),
                                                                                              kernelInfo(kernelInfo),
                                                                                              settings(settings) {
}


SgForStatement *NaiveCoarsener::buildCoarseningForLoop(SgStatement *body, int dim, SgScopeStatement *funcScope) {
    string ids[3] = {"coars_x", "coars_y", "coars_z"};
    string gids[3] = {"global_id_x", "global_id_y", "global_id_z"};

    int coarseningFactor = 0;
    if (dim == 0)
        coarseningFactor = params.elementsPerThreadX;
    if (dim == 1)
        coarseningFactor = params.elementsPerThreadY;
    if (dim == 2)
        coarseningFactor = params.elementsPerThreadZ;

    SgAssignInitializer *initialValue = buildAssignInitializer(buildIntVal(0));
    SgVariableDeclaration *init = buildVariableDeclaration(ids[dim], buildIntType(), initialValue, funcScope);

    SgVarRefExp *gidxTest = buildVarRefExp(gids[dim]);
    SgIntVal *cfTest = buildIntVal(coarseningFactor);
    SgAddOp *end = buildBinaryExpression<SgAddOp>(buildBinaryExpression<SgMultiplyOp>(gidxTest, cfTest), cfTest);
    SgExprStatement *test = buildExprStatement(
            buildLessThanOp(buildVarRefExp(ids[dim]), buildIntVal(coarseningFactor)));

    SgExpression *inc = buildPlusPlusOp(buildVarRefExp(ids[dim]), SgUnaryOp::postfix);

    SgBasicBlock *bodyBlock;
    if (isSgBasicBlock(body)) {
        bodyBlock = dynamic_cast<SgBasicBlock *>(body);
    } else {
        bodyBlock = buildBasicBlock();
        appendStatement(body, bodyBlock);
    }

    SgForStatement *forLoop = buildForStatement(init, test, inc, bodyBlock);

    return forLoop;
}


void NaiveCoarsener::removeReturn(SgBasicBlock *functionBody) {
    unsigned long nStatements = functionBody->get_statements().size();
    SgStatement *returnStatement = functionBody->get_statements().at(nStatements - 1);

    if (isSgReturnStmt(returnStatement)) {
        removeStatement(returnStatement);
    }
}


SgBasicBlock *NaiveCoarsener::getOriginalFunctionBody() {
    return originalFunctionBody;
}


void NaiveCoarsener::visit(SgNode *node) {
    SgFunctionDeclaration *functionDeclaration = isSgFunctionDeclaration(node);

    if (functionDeclaration != nullptr) {
        if (functionDeclaration->get_name() == kernelInfo.getKernelName()) {

            SgBasicBlock *functionBody = functionDeclaration->get_definition()->get_body();
            SgScopeStatement *functionScope = functionDeclaration->get_definition();
            SgScopeStatement *global = isSgScopeStatement(functionDeclaration->get_parent());

            removeReturn(functionBody);

            SgBasicBlock *loopBody = buildBasicBlock();

            moveStatementsBetweenBlocks(functionBody, loopBody);

            SgForStatement *xForLoop = buildCoarseningForLoop(loopBody, 0, functionBody);
            SgForStatement *yForLoop = buildCoarseningForLoop(xForLoop, 1, functionBody);
            SgForStatement *zForLoop = buildCoarseningForLoop(yForLoop, 2, functionBody);

            appendStatement(zForLoop, functionBody);

            SgVarRefExp *idx;
            SgVarRefExp *idy;
            SgVarRefExp *idz;
            if (settings.generateOMP || settings.generateMPI) {
                idx = buildVarRefExp("l_idx", functionScope);
                idy = buildVarRefExp("l_idy", functionScope);
                idz = buildVarRefExp("l_idz", functionScope);
            } else {
                idx = buildVarRefExp("idx", functionScope);
                idy = buildVarRefExp("idy", functionScope);
                idz = buildVarRefExp("idz", functionScope);
            }

            SgExpression *ggsX = buildOpaqueVarRefExp("GS_X", functionScope);
            SgExpression *ggsY = buildOpaqueVarRefExp("GS_Y", functionScope);
            SgExpression *ggsZ = buildOpaqueVarRefExp("GS_Z", functionScope);

            SgExpression *isOutside = nullptr;
            isOutside = buildOrOp(buildOrOp(buildGreaterOrEqualOp(idx, ggsX), buildGreaterOrEqualOp(idy, ggsY)), buildGreaterOrEqualOp(idz, ggsZ));

            SgIfStmt *ifOutside = buildIfStmt(isOutside, buildContinueStmt(), buildNullStatement());
            loopBody->prepend_statement(ifOutside);

            if (settings.generateMPI || settings.generateOMP) {

                SgAssignInitializer *bidx = buildAssignInitializer(buildAddOp(buildVarRefExp("l_idx", functionBody),
                                                                              buildOpaqueVarRefExp("base_x",
                                                                                                   functionScope)));
                SgVariableDeclaration *bidxDeclaration = buildVariableDeclaration("idx", buildIntType(), bidx,
                                                                                  functionBody);

                SgAssignInitializer *bidy = buildAssignInitializer(buildAddOp(buildVarRefExp("l_idy", functionBody),
                                                                              buildOpaqueVarRefExp("base_y",
                                                                                                   functionScope)));
                SgVariableDeclaration *bidyDeclaration = buildVariableDeclaration("idy", buildIntType(), bidy,
                                                                                  functionBody);

                SgAssignInitializer *bidz = buildAssignInitializer(buildAddOp(buildVarRefExp("l_idz", functionBody),
                                                                              buildOpaqueVarRefExp("base_z",
                                                                                                   functionScope)));
                SgVariableDeclaration *bidzDeclaration = buildVariableDeclaration("idz", buildIntType(), bidz,
                                                                                  functionBody);

                loopBody->prepend_statement(bidxDeclaration);
                loopBody->prepend_statement(bidyDeclaration);
                loopBody->prepend_statement(bidzDeclaration);
            }

            string idxName, idyName, idzName;
            if (settings.generateMPI || settings.generateOMP) {
                idxName = "l_idx";
                idyName = "l_idy";
                idzName = "l_idz";
            } else {
                idxName = "idx";
                idyName = "idy";
                idzName = "idz";
            }

            if (params.interleaved) {
                if (params.useLocalMem() || params.useRegCache()) {
                    if (settings.language == Language::OpenCL) {
                        // X
                        auto *getLocalIdx = buildFunctionCallExp("get_local_id", buildIntType(),
                                                                 buildExprListExp(buildIntVal(0)),
                                                                 functionBody);
                        auto *lidx = buildAddOp(getLocalIdx, buildMultiplyOp(buildIntVal(params.localSizeX),
                                                                             buildVarRefExp("coars_x",
                                                                                            functionBody)));
                        auto *getGroupIdx = buildFunctionCallExp("get_group_id", buildIntType(),
                                                                 buildExprListExp(buildIntVal(0)),
                                                                 functionBody);
                        auto *idxInit = buildAssignInitializer(buildAddOp(
                                buildMultiplyOp(getGroupIdx,
                                                buildIntVal(params.localSizeX * params.elementsPerThreadX)),
                                lidx));
                        auto *idxDecl = buildVariableDeclaration(idxName, buildIntType(), idxInit,
                                                                 functionBody);

                        // Y
                        auto *getLocalIdy = buildFunctionCallExp("get_local_id", buildIntType(),
                                                                 buildExprListExp(buildIntVal(1)),
                                                                 functionBody);
                        auto *lidy = buildAddOp(getLocalIdy, buildMultiplyOp(buildIntVal(params.localSizeY),
                                                                             buildVarRefExp("coars_y",
                                                                                            functionBody)));

                        auto *getGroupIdy = buildFunctionCallExp("get_group_id", buildIntType(),
                                                                 buildExprListExp(buildIntVal(1)),
                                                                 functionBody);
                        auto *idyInit = buildAssignInitializer(buildAddOp(
                                buildMultiplyOp(getGroupIdy,
                                                buildIntVal(params.localSizeY * params.elementsPerThreadY)),
                                lidy));
                        auto *idyDecl = buildVariableDeclaration(idyName, buildIntType(), idyInit,
                                                                 functionBody);

                        // Z
                        auto *getLocalIdz = buildFunctionCallExp("get_local_id", buildIntType(),
                                                                 buildExprListExp(buildIntVal(2)),
                                                                 functionBody);
                        auto *lidz = buildAddOp(getLocalIdz, buildMultiplyOp(buildIntVal(params.localSizeZ),
                                                                             buildVarRefExp("coars_z",
                                                                                            functionBody)));

                        auto *getGroupIdz = buildFunctionCallExp("get_group_id", buildIntType(),
                                                                 buildExprListExp(buildIntVal(2)),
                                                                 functionBody);
                        auto *idzInit = buildAssignInitializer(buildAddOp(
                                buildMultiplyOp(getGroupIdz,
                                                buildIntVal(params.localSizeZ * params.elementsPerThreadZ)),
                                lidz));
                        auto *idzDecl = buildVariableDeclaration(idzName, buildIntType(), idzInit,
                                                                 functionBody);

                        loopBody->prepend_statement(idzDecl);
                        loopBody->prepend_statement(idyDecl);
                        loopBody->prepend_statement(idxDecl);
                    } else if (settings.language == Language::Cuda) {
                        // X
                        auto *getThreadIdx = buildVarRefExp("threadIdx.x", functionBody);
                        auto *lidx = buildAddOp(getThreadIdx, buildMultiplyOp(buildIntVal(params.localSizeX),
                                                                              buildVarRefExp("coars_x", functionBody)));
                        auto *getBlockIdx = buildVarRefExp("blockIdx.x", functionBody);
                        auto *idxInit = buildAssignInitializer(buildAddOp(
                                buildMultiplyOp(getBlockIdx,
                                                buildIntVal(params.localSizeX * params.elementsPerThreadX)),
                                lidx
                        ));
                        auto *idxDecl = buildVariableDeclaration(idxName, buildIntType(), idxInit, functionBody);

                        // Y
                        auto *getThreadIdy = buildVarRefExp("threadIdx.y", functionBody);
                        auto *lidy = buildAddOp(getThreadIdy, buildMultiplyOp(buildIntVal(params.localSizeY),
                                                                              buildVarRefExp("coars_y", functionBody)));
                        auto *getBlockIdy = buildVarRefExp("blockIdx.y", functionBody);
                        auto *idyInit = buildAssignInitializer(buildAddOp(
                                buildMultiplyOp(getBlockIdy,
                                                buildIntVal(params.localSizeY * params.elementsPerThreadY)),
                                lidy
                        ));
                        auto *idyDecl = buildVariableDeclaration(idyName, buildIntType(), idyInit, functionBody);

                        // Z
                        auto *getThreadIdz = buildVarRefExp("threadIdx.z", functionBody);
                        auto *lidz = buildAddOp(getThreadIdz, buildMultiplyOp(buildIntVal(params.localSizeZ),
                                                                              buildVarRefExp("coars_z", functionBody)));
                        auto *getBlockIdz = buildVarRefExp("blockIdx.z", functionBody);
                        auto *idzInit = buildAssignInitializer(buildAddOp(
                                buildMultiplyOp(getBlockIdz,
                                                buildIntVal(params.localSizeZ * params.elementsPerThreadZ)),
                                lidz
                        ));
                        auto *idzDecl = buildVariableDeclaration(idzName, buildIntType(), idzInit, functionBody);

                        loopBody->prepend_statement(idzDecl);
                        loopBody->prepend_statement(idyDecl);
                        loopBody->prepend_statement(idxDecl);
                    }
                } else {
                    if (settings.language == Language::OpenCL) {
                        // X
                        auto *getGlobalIdx = buildFunctionCallExp("get_global_id", buildIntType(),
                                                                  buildExprListExp(buildIntVal(0)),
                                                                  functionBody);
                        auto *getGlobalSizeX = buildFunctionCallExp("get_global_size", buildIntType(),
                                                                    buildExprListExp(buildIntVal(0)),
                                                                    functionBody);
                        auto *idxInit = buildAssignInitializer(buildAddOp(getGlobalIdx,
                                                                          buildMultiplyOp(getGlobalSizeX,
                                                                                          buildVarRefExp(
                                                                                                  "coars_x",
                                                                                                  functionBody
                                                                                          )
                                                                          )
                                                               )
                        );
                        auto *idxDecl = buildVariableDeclaration(idxName, buildIntType(), idxInit, functionBody);

                        // Y
                        auto *getGlobalIdy = buildFunctionCallExp("get_global_id", buildIntType(),
                                                                  buildExprListExp(buildIntVal(1)),
                                                                  functionBody);
                        auto *getGlobalSizeY = buildFunctionCallExp("get_global_size", buildIntType(),
                                                                    buildExprListExp(buildIntVal(1)),
                                                                    functionBody);
                        auto *idyInit = buildAssignInitializer(buildAddOp(getGlobalIdy,
                                                                          buildMultiplyOp(getGlobalSizeY,
                                                                                          buildVarRefExp(
                                                                                                  "coars_y",
                                                                                                  functionBody
                                                                                          )
                                                                          )
                                                               )
                        );
                        auto *idyDecl = buildVariableDeclaration(idyName, buildIntType(), idyInit, functionBody);

                        // Z
                        auto *getGlobalIdz = buildFunctionCallExp("get_global_id", buildIntType(),
                                                                  buildExprListExp(buildIntVal(2)),
                                                                  functionBody);
                        auto *getGlobalSizeZ = buildFunctionCallExp("get_global_size", buildIntType(),
                                                                    buildExprListExp(buildIntVal(2)),
                                                                    functionBody);
                        auto *idzInit = buildAssignInitializer(buildAddOp(getGlobalIdz,
                                                                          buildMultiplyOp(getGlobalSizeZ,
                                                                                          buildVarRefExp(
                                                                                                  "coars_z",
                                                                                                  functionBody
                                                                                          )
                                                                          )
                                                               )
                        );
                        auto *idzDecl = buildVariableDeclaration(idzName, buildIntType(), idzInit, functionBody);
                        loopBody->prepend_statement(idzDecl);
                        loopBody->prepend_statement(idyDecl);
                        loopBody->prepend_statement(idxDecl);
                    } else if (settings.language == Language::Cuda) {
                        // X
                        // threadIdx.x + blockIdx.x * blockDim.x
                        auto *getGlobalIdx = buildAddOp(
                                buildVarRefExp("threadIdx.x", functionBody),
                                buildMultiplyOp(buildVarRefExp("blockIdx.x", functionBody),
                                                buildVarRefExp("blockDim.x", functionBody))
                        );
                        // (threadIdx.x + blockIdx.x * blockDim.x) + gridDim.x * blockDim.x * coars_x
                        getGlobalIdx = buildAddOp(
                                getGlobalIdx,
                                buildMultiplyOp(
                                        buildMultiplyOp(buildVarRefExp("gridDim.x", functionBody),
                                                        buildVarRefExp("blockDim.x", functionBody)),
                                        buildVarRefExp("coars_x", functionBody)
                                )
                        );
                        auto *idxInit = buildAssignInitializer(getGlobalIdx);
                        auto *idxDecl = buildVariableDeclaration(idxName, buildIntType(), idxInit, functionBody);

                        // Y
                        // threadIdx.y + blockIdx.y * blockDim.y
                        auto *getGlobalIdy = buildAddOp(
                                buildVarRefExp("threadIdx.y", functionBody),
                                buildMultiplyOp(buildVarRefExp("blockIdx.y", functionBody),
                                                buildVarRefExp("blockDim.y", functionBody))
                        );
                        // (threadIdx.y + blockIdx.y * blockDim.y) + gridDim.y * blockDim.y * coars_y
                        getGlobalIdy = buildAddOp(
                                getGlobalIdy,
                                buildMultiplyOp(
                                        buildMultiplyOp(buildVarRefExp("gridDim.y", functionBody),
                                                        buildVarRefExp("blockDim.y", functionBody)),
                                        buildVarRefExp("coars_y", functionBody)
                                )
                        );
                        auto *idyInit = buildAssignInitializer(getGlobalIdy);
                        auto *idyDecl = buildVariableDeclaration(idyName, buildIntType(), idyInit, functionBody);

                        // Z
                        // threadIdx.z + blockIdx.z * blockDim.z
                        auto *getGlobalIdz = buildAddOp(
                                buildVarRefExp("threadIdx.z", functionBody),
                                buildMultiplyOp(buildVarRefExp("blockIdx.z", functionBody),
                                                buildVarRefExp("blockDim.z", functionBody))
                        );
                        // (threadIdx.z + blockIdx.z * blockDim.z) + gridDim.z * blockDim.z * coars_z
                        getGlobalIdz = buildAddOp(
                                getGlobalIdz,
                                buildMultiplyOp(
                                        buildMultiplyOp(buildVarRefExp("gridDim.z", functionBody),
                                                        buildVarRefExp("blockDim.z", functionBody)),
                                        buildVarRefExp("coars_z", functionBody)
                                )
                        );
                        auto *idzInit = buildAssignInitializer(getGlobalIdz);
                        auto *idzDecl = buildVariableDeclaration(idzName, buildIntType(), idzInit, functionBody);

                        loopBody->prepend_statement(idzDecl);
                        loopBody->prepend_statement(idyDecl);
                        loopBody->prepend_statement(idxDecl);
                    }
                }
            } else {
                if (settings.language == Language::OpenCL) {
                    // X
                    auto *getGlobalIdx = buildFunctionCallExp("get_global_id", buildIntType(),
                                                              buildExprListExp(buildIntVal(0)),
                                                              functionBody);
                    auto *idxInit = buildAssignInitializer(
                            buildAddOp(buildMultiplyOp(getGlobalIdx, buildIntVal(params.elementsPerThreadX)),
                                       buildVarRefExp("coars_x", functionBody)));
                    auto *idxDecl = buildVariableDeclaration(idxName, buildIntType(), idxInit,
                                                             functionBody);

                    // Y
                    auto *getGlobalIdy = buildFunctionCallExp("get_global_id", buildIntType(),
                                                              buildExprListExp(buildIntVal(1)),
                                                              functionBody);
                    auto *idyInit = buildAssignInitializer(
                            buildAddOp(buildMultiplyOp(getGlobalIdy, buildIntVal(params.elementsPerThreadY)),
                                       buildVarRefExp("coars_y", functionBody)));
                    auto *idyDecl = buildVariableDeclaration(idyName, buildIntType(), idyInit,
                                                             functionBody);

                    // Z
                    auto *getGlobalIdz = buildFunctionCallExp("get_global_id", buildIntType(),
                                                              buildExprListExp(buildIntVal(2)),
                                                              functionBody);
                    auto *idzInit = buildAssignInitializer(
                            buildAddOp(buildMultiplyOp(getGlobalIdz, buildIntVal(params.elementsPerThreadZ)),
                                       buildVarRefExp("coars_z", functionBody)));
                    auto *idzDecl = buildVariableDeclaration(idzName, buildIntType(), idzInit,
                                                             functionBody);

                    loopBody->prepend_statement(idzDecl);
                    loopBody->prepend_statement(idyDecl);
                    loopBody->prepend_statement(idxDecl);
                } else if (settings.language == Language::Cuda) {
                    // X
                    auto *getGlobalIdx = buildAddOp(
                            buildVarRefExp("threadIdx.x", functionBody),
                            buildMultiplyOp(buildVarRefExp("blockIdx.x", functionBody),
                                            buildVarRefExp("blockDim.x", functionBody))
                    );
                    auto *idxInit = buildAssignInitializer(
                            buildAddOp(buildMultiplyOp(getGlobalIdx, buildIntVal(params.elementsPerThreadX)),
                                       buildVarRefExp("coars_x", functionBody))
                    );
                    auto *idxDecl = buildVariableDeclaration(idxName, buildIntType(), idxInit, functionBody);

                    // Y
                    auto *getGlobalIdy = buildAddOp(
                            buildVarRefExp("threadIdx.y", functionBody),
                            buildMultiplyOp(buildVarRefExp("blockIdx.y", functionBody),
                                            buildVarRefExp("blockDim.y", functionBody))
                    );
                    auto *idyInit = buildAssignInitializer(
                            buildAddOp(buildMultiplyOp(getGlobalIdy, buildIntVal(params.elementsPerThreadY)),
                                       buildVarRefExp("coars_y", functionBody))
                    );
                    auto *idyDecl = buildVariableDeclaration(idyName, buildIntType(), idyInit, functionBody);

                    // Z
                    auto *getGlobalIdz = buildAddOp(
                            buildVarRefExp("threadIdx.z", functionBody),
                            buildMultiplyOp(buildVarRefExp("blockIdx.z", functionBody),
                                            buildVarRefExp("blockDim.z", functionBody))
                    );
                    auto *idzInit = buildAssignInitializer(
                            buildAddOp(buildMultiplyOp(getGlobalIdz, buildIntVal(params.elementsPerThreadZ)),
                                       buildVarRefExp("coars_z", functionBody))
                    );
                    auto *idzDecl = buildVariableDeclaration(idzName, buildIntType(), idzInit, functionBody);

                    loopBody->prepend_statement(idzDecl);
                    loopBody->prepend_statement(idyDecl);
                    loopBody->prepend_statement(idxDecl);
                }
            }
            originalFunctionBody = loopBody;
        }

        // fixVariableReferences(functionDeclaration);
    }
}
