#ifndef MINIAMR_GPU_H
#define MINIAMR_GPU_H

#if defined(MINIAMR_CUDA) || defined(MINIAMR_OPENCL)

#include "block.h"
#include "proto.h"
#include "timer.h"

#include <cuda.h>
#include <cuda_runtime.h>
#include <stdio.h>
#include <vector_types.h>

#define ERR(e) if ((e) != cudaSuccess) { fprintf(stderr, "err on line %d: %d -> %s\n", __LINE__, cudaGetLastError(), cudaGetErrorString(cudaGetLastError())); exit(-1); }
#define X (x_block_size + 2)
#define Y (y_block_size + 2)
#define Z (z_block_size + 2)

struct gpu_params init_gpu(int start, int local_num_vars) {
    int num_iter = sorted_index[num_refine + 1];

#if defined(V2) || defined(V3) || defined(V5) || defined(V6)
    int data_size = num_iter * local_num_vars * Z * Y * X;
#else
    // We allocate one extra array for the variables so we can shift it in the stencil code.
    int data_size = num_iter * (local_num_vars + 1) * Z * Y * X;
#endif
    double *data = ma_malloc(data_size * sizeof(double), __FILE__, __LINE__);

    // Allocate on the GPU.
    double *data_gpu;
    ERR(ma_cudaMalloc((void**) &data_gpu, data_size * sizeof(double)));

#if defined(V2) || defined(V3) || defined(V4) || defined(V5) || defined(V6)
    double *output_gpu;
    ERR(ma_cudaMalloc((void**) &output_gpu, data_size * sizeof(double)));
#endif

#if defined(V6)
    // Since double is not supported in texture memory, we need to use int2 and cast it to double.
    // See: https://developer.nvidia.com/cuda-faq
    struct cudaChannelFormatDesc channelDesc_input = cudaCreateChannelDesc(32, 32, 0, 0, cudaChannelFormatKindSigned);
    cudaArray_t cuda_array;
    ERR(ma_cudaMalloc3DArray(&cuda_array, &channelDesc_input, make_cudaExtent((size_t) Z, (size_t) Y, (size_t) X)));
#endif

    struct gpu_params params = {
            .num_iter = num_iter,
            .num_vars = local_num_vars,
            .data_size = data_size,
            .data = data,
            .data_gpu = data_gpu,
#if defined(V2) || defined(V3) || defined(V4) || defined(V5) || defined(V6)
            .output_gpu = output_gpu,
            .start = start,
#endif
#if defined(V6)
            .cuda_array = cuda_array,
#endif
    };

#if defined(V2) || defined(V3) || defined(V4) || defined(V5) || defined(V6)
    for (int i = 0; i < num_iter; ++i) {
        for (int j = 0; j < local_num_vars; ++j) {
            for (int x = 0; x < X; ++x) {
                for (int y = 0; y < Y; ++y) {
                    for (int z = 0; z < Z; ++z) {
                        params.data[
                                (i * local_num_vars * X * Y * Z) +
                                (j * X * Y * Z) +
                                (x * Y * Z) +
                                (y * Z) +
                                z
                        ] = blocks[sorted_list[i].n].array[j + start][x][y][z];
                    }
                }
            }
        }
    }

    // Transfer data onto the GPU.
    ERR(ma_cudaMemcpy(start % 2 == 0 ? params.data_gpu : params.output_gpu,
                   params.data,
                   params.data_size * sizeof(double),
                   cudaMemcpyHostToDevice));
    ERR(ma_cudaMemcpy(start % 2 == 0 ? params.output_gpu : params.data_gpu,
                   start % 2 == 0 ? params.data_gpu : params.output_gpu,
                   params.data_size * sizeof(double),
                   cudaMemcpyDeviceToDevice));
#else
    for (int i = 0; i < num_iter; ++i) {
        for (int j = 1; j <= local_num_vars; ++j) {
            for (int x = 0; x < X; ++x) {
                for (int y = 0; y < Y; ++y) {
                    for (int z = 0; z < Z; ++z) {
                        params.data[
                             (i * (local_num_vars + 1) * X * Y * Z) +
                             (j * X * Y * Z) +
                             (x * Y * Z) +
                             (y * Z) +
                             z
                        ] = blocks[sorted_list[i].n].array[j + start - 1][x][y][z];
                    }
                }
            }
        }
    }

    // Transfer data onto the GPU.
    ERR(ma_cudaMemcpy(params.data_gpu, params.data, params.data_size * sizeof(double), cudaMemcpyHostToDevice));
#endif

    return params;
}

void free_gpu(struct gpu_params *params, int start, int local_num_vars) {
#if defined(V2) || defined(V3) || defined(V4) || defined(V5) || defined(V6)
    ERR(ma_cudaMemcpy(params->data, (start + local_num_vars - 1) % 2 == 0 ? params->data_gpu: params->output_gpu, params->data_size * sizeof(double), cudaMemcpyDeviceToHost));

    for (int i = 0; i < params->num_iter; ++i) {
        for (int j = 0; j < local_num_vars; ++j) {
            for (int x = 0; x < X; ++x) {
                for (int y = 0; y < Y; ++y) {
                    for (int z = 0; z < Z; ++z) {
                        blocks[sorted_list[i].n].array[j + start][x][y][z] = params->data[
                                (i * local_num_vars * X * Y * Z) +
                                (j * X * Y * Z) +
                                (x * Y * Z) +
                                (y * Z) +
                                z
                        ];
                    }
                }
            }
        }
    }
#else
    ERR(ma_cudaMemcpy(params->data, params->data_gpu, params->data_size * sizeof(double), cudaMemcpyDeviceToHost));

    for (int i = 0; i < params->num_iter; ++i) {
        for (int j = 0; j < local_num_vars; ++j) {
            for (int x = 0; x < X; ++x) {
                for (int y = 0; y < Y; ++y) {
                    for (int z = 0; z < Z; ++z) {
                        blocks[sorted_list[i].n].array[j + start][x][y][z] = params->data[
                                (i * (local_num_vars + 1) * X * Y * Z) +
                                (j * X * Y * Z) +
                                (x * Y * Z) +
                                (y * Z) +
                                z
                        ];
                    }
                }
            }
        }
    }
#endif

    free(params->data);
    params->data = NULL;
    ERR(cudaFree(params->data_gpu));
    params->data_gpu = NULL;
#if defined(V2) || defined(V3) || defined(V4) || defined(V5) || defined(V6)
    ERR(cudaFree(params->output_gpu));
    params->output_gpu = NULL;
#endif
#if defined(V6)
    ERR(cudaFreeArray(params->cuda_array));
    params->cuda_array = NULL;
#endif
}

#endif

#endif //MINIAMR_GPU_H
