#include <stdio.h>
#include <cuda.h>
#include <driver_types.h>
#include <cuda_runtime.h>
#include <device_launch_parameters.h>
#include <memory.h>

extern "C" struct gpu_params {
    int num_iter;
    int num_vars;
    int data_size;
    double *data;
    double *data_gpu;
#if defined(V2) || defined(V3) || defined(V4) || defined(V5) || defined(V6)
    double *output_gpu;
    int start;
#endif
#if defined(V6)
    cudaArray_t cuda_array;
#endif
};

extern "C" double kernel_event_time;
extern "C" size_t num_kernel_calls;
extern "C" double largest_diff_with_events;
extern "C" void stencil_calc(int, struct gpu_params *);
extern "C" double check_sum_calc(int, struct gpu_params *);
extern "C" cudaError_t ma_cudaMalloc(void **devPtr, size_t size);
extern "C" cudaError_t ma_cudaMemcpy(void *dst, const void *src, size_t count, enum cudaMemcpyKind kind);
extern "C" cudaError_t ma_cudaMemcpy3D(const struct cudaMemcpy3DParms *p);
extern "C" cudaError_t ma_cudaMemcpyToSymbol(const void *symbol, const void *src, size_t count, size_t offset, enum cudaMemcpyKind kind);

constexpr int XG = 8;
constexpr int YG = 8;
constexpr int ZG = 8;
constexpr int BLOCK_X = 8;
constexpr int BLOCK_Y = 8;
constexpr int EPT_X = 1;
constexpr int EPT_Y = 1;
#if defined(V1) || defined(V2) || defined(V3) || defined(V4) || defined(V5) || defined(V6)
constexpr int BLOCK_Z = 8;
constexpr int EPT_Z = 1;
#endif

#define CHECK_ERROR(e) if ((e) != cudaSuccess) { fprintf(stderr, "err on line %d: %d -> %s\n", __LINE__, cudaGetLastError(), cudaGetErrorString(cudaGetLastError())); exit(90); }

#if defined(V0)
/// miniAMR v0 is the default 2d version and writes and reads to same memory with an offset to avoid block synchronization.
///
/// This version can only be used with global memory, but has hopefully a more efficient memory access.
__global__ void miniAMR(int var, int num_iter, int num_vars, double *data, int GS_X, int GS_Y) {
    #pragma unroll
    for (int coars_y = 0; coars_y < EPT_Y; coars_y++) {
        #pragma unroll
        for (int coars_x = 0; coars_x < EPT_X; coars_x++) {
            int idx = threadIdx.x + blockIdx.x * blockDim.x + gridDim.x * blockDim.x * coars_x;
            int idy = threadIdx.y + blockIdx.y * blockDim.y + gridDim.y * blockDim.y * coars_y;
            if (idx >= GS_X || idy >= GS_Y)
                continue;
            else;
            constexpr int X = XG + 2;
            constexpr int Y = YG + 2;
            constexpr int Z = ZG + 2;
            int idx1 = idx + 1;
            int idy1 = idy + 1;

            #pragma unroll
            for (int i = 0; i < num_iter; ++i) {
                // 5d -> 3d
                double *read = &data[i * (num_vars + 1) * X * Y * Z + (var + 1) * X * Y * Z];
                double *write = &data[i * (num_vars + 1) * X * Y * Z + var * X * Y * Z];
                for (int z = 1; z <= Z - 2; ++z) {
                    write[idx1 * Y * Z + idy1 * Z + z] =
                            (read[idx1 * Y * Z + idy1 * Z + (z - 1)]
                             + read[idx1 * Y * Z + (idy1 - 1) * Z + z]
                             + read[(idx1 - 1) * Y * Z + idy1 * Z + z]
                             + read[idx1 * Y * Z + idy1 * Z + z]
                             + read[(idx1 + 1) * Y * Z + idy1 * Z + z]
                             + read[idx1 * Y * Z + (idy1 + 1) * Z + z]
                             + read[idx1 * Y * Z + idy1 * Z + (z + 1)]) / 7.0;
                }
            }
        }
    }
}

#elif defined(V1)
/// miniAMR v1 writes and reads to same memory with an offset to avoid block synchronization.
///
/// This version can only be used with global memory, but has hopefully a more efficient memory access.
__global__ void
miniAMR(int var, int num_iter, int num_vars, double *data, int data_width, int data_height, int data_depth,
           int GS_X, int GS_Y, int GS_Z) {
    #pragma unroll
    for (int coars_z = 0; coars_z < EPT_Z; coars_z++) {
        #pragma unroll
        for (int coars_y = 0; coars_y < EPT_Y; coars_y++) {
            #pragma unroll
            for (int coars_x = 0; coars_x < EPT_X; coars_x++) {
                int idx = threadIdx.x + blockIdx.x * blockDim.x + gridDim.x * blockDim.x * coars_x;
                int idy = threadIdx.y + blockIdx.y * blockDim.y + gridDim.y * blockDim.y * coars_y;
                int idz = threadIdx.z + blockIdx.z * blockDim.z + gridDim.z * blockDim.z * coars_z;
                if (idx >= GS_X || idy >= GS_Y || idz >= GS_Z)
                    continue;
                else;
                #pragma unroll
                for (int i = 0; i < num_iter; ++i) {
                    const int READ = (i * (num_vars + 1) + (var + 1)) * (XG + 2) * (YG + 2) * (ZG + 2);
                    const int WRITE = (i * (num_vars + 1) + var) * (XG + 2) * (YG + 2) * (ZG + 2);
                    // 5d -> 3d
                    double *write = data + WRITE;
                    double *read = data + READ;
                    write[(idx + 1) * (data_width * data_height) + ((idy + 1) * data_width + (idz + 1))] =
                            (read[(idx + 1) * (data_width * data_height) + ((idy + 1) * data_width + idz)] +
                             read[(idx + 1) * (data_width * data_height) + (idy * data_width + (idz + 1))] +
                             read[idx * (data_width * data_height) + ((idy + 1) * data_width + (idz + 1))] +
                             read[(idx + 1) * (data_width * data_height) + ((idy + 1) * data_width + (idz + 1))] +
                             read[(idx + 2) * (data_width * data_height) + ((idy + 1) * data_width + (idz + 1))] +
                             read[(idx + 1) * (data_width * data_height) + ((idy + 2) * data_width + (idz + 1))] +
                             read[(idx + 1) * (data_width * data_height) + ((idy + 1) * data_width + (idz + 2))]) / 7.0;
                }
            }
        }
    }
}
#elif defined(V2)
/// miniAMR v2 uses separate memory for reading and writing to avoid block synchronization.
///
/// This version can be used together with global and hopefully constant memory.
__global__ void
miniAMR(int var, int num_iter, int num_vars, double *input, double *output, int input_width, int input_height,
           int input_depth, int output_width, int output_height, int output_depth, int GS_X, int GS_Y, int GS_Z) {
    #pragma unroll
    for (int coars_z = 0; coars_z < EPT_Z; coars_z++) {
        #pragma unroll
        for (int coars_y = 0; coars_y < EPT_Y; coars_y++) {
            #pragma unroll
            for (int coars_x = 0; coars_x < EPT_X; coars_x++) {
                int idx = threadIdx.x + blockIdx.x * blockDim.x + gridDim.x * blockDim.x * coars_x;
                int idy = threadIdx.y + blockIdx.y * blockDim.y + gridDim.y * blockDim.y * coars_y;
                int idz = threadIdx.z + blockIdx.z * blockDim.z + gridDim.z * blockDim.z * coars_z;
                if (idx >= GS_X || idy >= GS_Y || idz >= GS_Z)
                    continue;
                else;
                #pragma unroll
                for (int i = 0; i < num_iter; ++i) {
                    const int OFFSET = (i * num_vars + var) * (XG + 2) * (YG + 2) * (ZG + 2);
                    // 5d -> 3d
                    double *read = input + OFFSET;
                    double *write = output + OFFSET;
                    write[(idx + 1) * (output_width * output_height) + ((idy + 1) * output_width + (idz + 1))] =
                            (read[(idx + 1) * (input_width * input_height) + ((idy + 1) * input_width + idz)] +
                             read[(idx + 1) * (input_width * input_height) + (idy * input_width + (idz + 1))] +
                             read[idx * (input_width * input_height) + ((idy + 1) * input_width + (idz + 1))] +
                             read[(idx + 1) * (input_width * input_height) + ((idy + 1) * input_width + (idz + 1))] +
                             read[(idx + 2) * (input_width * input_height) + ((idy + 1) * input_width + (idz + 1))] +
                             read[(idx + 1) * (input_width * input_height) + ((idy + 2) * input_width + (idz + 1))] +
                             read[(idx + 1) * (input_width * input_height) + ((idy + 1) * input_width + (idz + 2))]) / 7.0;
                }
            }
        }
    }
}
#elif defined(V3)
/// miniAMR v3 uses 3d memory instead of 5d.
///
/// This version can hopefully be used together with all memory optimizations.
__global__ void
miniAMR(double *input, double *output, int input_width, int input_height, int input_depth, int output_width,
               int output_height, int output_depth, int GS_X, int GS_Y, int GS_Z) {
    #pragma unroll
    for (int coars_z = 0; coars_z < EPT_Z; coars_z++) {
        #pragma unroll
        for (int coars_y = 0; coars_y < EPT_Y; coars_y++) {
            #pragma unroll
            for (int coars_x = 0; coars_x < EPT_X; coars_x++) {
                int idx = threadIdx.x + blockIdx.x * blockDim.x + gridDim.x * blockDim.x * coars_x;
                int idy = threadIdx.y + blockIdx.y * blockDim.y + gridDim.y * blockDim.y * coars_y;
                int idz = threadIdx.z + blockIdx.z * blockDim.z + gridDim.z * blockDim.z * coars_z;
                if (idx >= GS_X || idy >= GS_Y || idz >= GS_Z)
                    continue;
                else;
                output[(idx + 1) * (output_width * output_height) + ((idy + 1) * output_width + (idz + 1))] =
                        (input[(idx + 1) * (input_width * input_height) + ((idy + 1) * input_width + idz)] +
                         input[(idx + 1) * (input_width * input_height) + (idy * input_width + (idz + 1))] +
                         input[idx * (input_width * input_height) + ((idy + 1) * input_width + (idz + 1))] +
                         input[(idx + 1) * (input_width * input_height) + ((idy + 1) * input_width + (idz + 1))] +
                         input[(idx + 2) * (input_width * input_height) + ((idy + 1) * input_width + (idz + 1))] +
                         input[(idx + 1) * (input_width * input_height) + ((idy + 2) * input_width + (idz + 1))] +
                         input[(idx + 1) * (input_width * input_height) + ((idy + 1) * input_width + (idz + 2))]) / 7.0;
            }
        }
    }
}
#elif defined(V4)
/// V3 with shared memory.
__global__ void
miniAMR(double *input, double *output, int input_width, int input_height, int input_depth, int output_width,
        int output_height, int output_depth, int GS_X, int GS_Y, int GS_Z) {
    __device__ __shared__ double local_buffer_input[BLOCK_X + 2][BLOCK_Y + 2][BLOCK_Z + 2];
    int smem_load_input = threadIdx.x + (threadIdx.y * blockDim.x + threadIdx.z * (blockDim.x * blockDim.y));
    #pragma unroll
    while (smem_load_input < ((BLOCK_X + 2) * (BLOCK_Y + 2) * (BLOCK_Z + 2))) {
        const int local_x_input = smem_load_input % ((BLOCK_X + 2) * (BLOCK_Y + 2)) % (BLOCK_X + 2);
        const int local_y_input = smem_load_input % ((BLOCK_X + 2) * (BLOCK_Y + 2)) / (BLOCK_X + 2);
        const int local_z_input = smem_load_input / ((BLOCK_X + 2) * (BLOCK_Y + 2));
        const int global_x_input = blockIdx.x * BLOCK_X + local_x_input - 0;
        const int global_y_input = blockIdx.y * BLOCK_Y + local_y_input - 0;
        const int global_z_input = blockIdx.z * BLOCK_Z + local_z_input - 0;
        local_buffer_input[local_x_input][local_y_input][local_z_input] = input[
                global_x_input * (input_width * input_height) + (global_y_input * input_width + global_z_input)];
        smem_load_input += BLOCK_X * BLOCK_Y * BLOCK_Z;
    }
    __syncthreads();
    #pragma unroll
    for (int coars_z = 0; coars_z < EPT_Z; coars_z++) {
        #pragma unroll
        for (int coars_y = 0; coars_y < EPT_Y; coars_y++) {
            #pragma unroll
            for (int coars_x = 0; coars_x < EPT_X; coars_x++) {
                int lidx_input = threadIdx.x + BLOCK_X * coars_x + 0;
                int lidy_input = threadIdx.y + BLOCK_Y * coars_y + 0;
                int lidz_input = threadIdx.z + BLOCK_Z * coars_z + 0;
                int idx = threadIdx.x + blockIdx.x * blockDim.x + gridDim.x * blockDim.x * coars_x;
                int idy = threadIdx.y + blockIdx.y * blockDim.y + gridDim.y * blockDim.y * coars_y;
                int idz = threadIdx.z + blockIdx.z * blockDim.z + gridDim.z * blockDim.z * coars_z;
                if (idx >= GS_X || idy >= GS_Y || idz >= GS_Z)
                    continue;
                else;
                output[(idx + 1) * (output_width * output_height) + ((idy + 1) * output_width + (idz + 1))] =
                        (local_buffer_input[lidx_input][lidy_input + 1][lidz_input + 1] +
                         local_buffer_input[lidx_input + 1][lidy_input][lidz_input + 1] +
                         local_buffer_input[lidx_input + 1][lidy_input + 1][lidz_input] +
                         local_buffer_input[lidx_input + 1][lidy_input + 1][lidz_input + 1] +
                         local_buffer_input[lidx_input + 1][lidy_input + 1][lidz_input + 2] +
                         local_buffer_input[lidx_input + 1][lidy_input + 2][lidz_input + 1] +
                         local_buffer_input[lidx_input + 2][lidy_input + 1][lidz_input + 1]) / 7.0;
            }
        }
    }
}
#elif defined(V5)
/// V3 with constant memory.
__constant__ double const_input[(XG + 2) * (YG + 2) * (ZG + 2)];

__global__ void
miniAMR(double *output, int input_width, int input_height, int input_depth, int output_width, int output_height,
        int output_depth, int GS_X, int GS_Y, int GS_Z) {
    #pragma unroll
    for (int coars_z = 0; coars_z < EPT_Z; coars_z++) {
        #pragma unroll
        for (int coars_y = 0; coars_y < EPT_Y; coars_y++) {
            #pragma unroll
            for (int coars_x = 0; coars_x < EPT_X; coars_x++) {
                int idx = threadIdx.x + blockIdx.x * blockDim.x + gridDim.x * blockDim.x * coars_x;
                int idy = threadIdx.y + blockIdx.y * blockDim.y + gridDim.y * blockDim.y * coars_y;
                int idz = threadIdx.z + blockIdx.z * blockDim.z + gridDim.z * blockDim.z * coars_z;
                if (idx >= GS_X || idy >= GS_Y || idz >= GS_Z)
                    continue;
                else;
                output[(idx + 1) * (output_width * output_height) + ((idy + 1) * output_width + (idz + 1))] =
                        (const_input[(idx + 1) * (input_width * input_height) + ((idy + 1) * input_width + idz)] +
                         const_input[(idx + 1) * (input_width * input_height) + (idy * input_width + (idz + 1))] +
                         const_input[idx * (input_width * input_height) + ((idy + 1) * input_width + (idz + 1))] +
                         const_input[(idx + 1) * (input_width * input_height) + ((idy + 1) * input_width + (idz + 1))] +
                         const_input[(idx + 2) * (input_width * input_height) + ((idy + 1) * input_width + (idz + 1))] +
                         const_input[(idx + 1) * (input_width * input_height) + ((idy + 2) * input_width + (idz + 1))] +
                         const_input[(idx + 1) * (input_width * input_height) + ((idy + 1) * input_width + (idz + 2))]) / 7.0;
            }
        }
    }
}
#elif defined(V6)
/// V3 with texture memory.
__global__ void
miniAMR(cudaTextureObject_t input, double *output, int input_width, int input_height, int input_depth, int output_width,
        int output_height, int output_depth, int GS_X, int GS_Y, int GS_Z) {
    #pragma unroll
    for (int coars_z = 0; coars_z < EPT_Z; coars_z++) {
        #pragma unroll
        for (int coars_y = 0; coars_y < EPT_Y; coars_y++) {
            #pragma unroll
            for (int coars_x = 0; coars_x < EPT_X; coars_x++) {
                int idx = threadIdx.x + blockIdx.x * blockDim.x + gridDim.x * blockDim.x * coars_x;
                int idy = threadIdx.y + blockIdx.y * blockDim.y + gridDim.y * blockDim.y * coars_y;
                int idz = threadIdx.z + blockIdx.z * blockDim.z + gridDim.z * blockDim.z * coars_z;
                if (idx >= GS_X || idy >= GS_Y || idz >= GS_Z)
                    continue;
                else;
                int2 a = tex3D<int2>(input, idz, idy + 1, idx + 1);
                int2 b = tex3D<int2>(input, idz + 1, idy, idx + 1);
                int2 c = tex3D<int2>(input, idz + 1, idy + 1, idx);
                int2 d = tex3D<int2>(input, idz + 1, idy + 1, idx + 1);
                int2 e = tex3D<int2>(input, idz + 1, idy + 1, idx + 2);
                int2 f = tex3D<int2>(input, idz + 1, idy + 2, idx + 1);
                int2 g = tex3D<int2>(input, idz + 2, idy + 1, idx + 1);
                output[(idx + 1) * (output_width * output_height) + ((idy + 1) * output_width + (idz + 1))] =
                        (__hiloint2double(a.y, a.x) +
                         __hiloint2double(b.y, b.x) +
                         __hiloint2double(c.y, c.x) +
                         __hiloint2double(d.y, d.x) +
                         __hiloint2double(e.y, e.x) +
                         __hiloint2double(f.y, f.x) +
                         __hiloint2double(g.y, g.x)) / 7.0;
            }
        }
    }
}
#else
#error "cuda version not defined"
#endif

void stencil_calc(int var, struct gpu_params *params) {
    // We initialize and de-initialize memory in the init_gpu and free_gpu methods.
    int gsx = XG;
    int gsy = YG;
#if defined(V1) || defined(V2) || defined(V3) || defined(V4) || defined(V5) || defined(V6)
    int gsz = ZG;
#endif

    // Timing
    cudaEvent_t start, stop;
    cudaEventCreate(&start);
    cudaEventCreate(&stop);

#if defined(V0)
    dim3 block(BLOCK_X, BLOCK_Y);
    dim3 grid((gsx + BLOCK_X * EPT_X - 1) / (BLOCK_X * EPT_X), (gsy + BLOCK_Y * EPT_Y - 1) / (BLOCK_Y * EPT_Y));
    cudaEventRecord(start);
    miniAMR<<<grid, block>>>(
            var,
            params->num_iter,
            params->num_vars,
            params->data_gpu,
            gsx,
            gsy);
    cudaEventRecord(stop);
    CHECK_ERROR(cudaEventSynchronize(stop));
    float ms = 0.0;
    CHECK_ERROR(cudaEventElapsedTime(&ms, start, stop));
    kernel_event_time += (double) ms;

    if (ms > largest_diff_with_events) {
        largest_diff_with_events = ms;
    }

    num_kernel_calls += 1;
#elif defined(V1)
    dim3 block(BLOCK_X, BLOCK_Y, BLOCK_Z);
    dim3 grid((gsx + BLOCK_X * EPT_X - 1) / (BLOCK_X * EPT_X), (gsy + BLOCK_Y * EPT_Y - 1) / (BLOCK_Y * EPT_Y), (gsz + BLOCK_Z * EPT_Z - 1) / (BLOCK_Z * EPT_Z));
    cudaEventRecord(start);
    miniAMR<<<grid, block>>>(
            var,
            params->num_iter,
            params->num_vars,
            params->data_gpu,
            gsz + 2,
            gsy + 2,
            gsx + 2,
            gsx,
            gsy,
            gsz);
    cudaEventRecord(stop);

    CHECK_ERROR(cudaEventSynchronize(stop));
    float ms = 0.0;
    CHECK_ERROR(cudaEventElapsedTime(&ms, start, stop));
    kernel_event_time += (double) ms;
    if (ms > largest_diff_with_events) {
        largest_diff_with_events = ms;
    }

    num_kernel_calls += 1;
#elif defined(V2)
    dim3 block(BLOCK_X, BLOCK_Y, BLOCK_Z);
    dim3 grid((gsx + BLOCK_X * EPT_X - 1) / (BLOCK_X * EPT_X), (gsy + BLOCK_Y * EPT_Y - 1) / (BLOCK_Y * EPT_Y), (gsz + BLOCK_Z * EPT_Z - 1) / (BLOCK_Z * EPT_Z));
    cudaEventRecord(start);
    miniAMR<<<grid, block>>>(
            var,
            params->num_iter,
            params->num_vars,
            (var - params->start) % 2 == 0 ? params->data_gpu : params->output_gpu,
            (var - params->start) % 2 == 0 ? params->output_gpu : params->data_gpu,
            gsz + 2,
            gsy + 2,
            gsx + 2,
            gsz + 2,
            gsy + 2,
            gsx + 2,
            gsx,
            gsy,
            gsz);
    cudaEventRecord(stop);

    CHECK_ERROR(cudaEventSynchronize(stop));
    float ms = 0.0;
    CHECK_ERROR(cudaEventElapsedTime(&ms, start, stop));
    kernel_event_time += (double) ms;
    if (ms > largest_diff_with_events) {
        largest_diff_with_events = ms;
    }

    num_kernel_calls += 1;
#elif defined(V3) || defined(V4)
    dim3 block(BLOCK_X, BLOCK_Y, BLOCK_Z);
    dim3 grid((gsx + BLOCK_X * EPT_X - 1) / (BLOCK_X * EPT_X),
              (gsy + BLOCK_Y * EPT_Y - 1) / (BLOCK_Y * EPT_Y),
              (gsz + BLOCK_Z * EPT_Z - 1) / (BLOCK_Z * EPT_Z));
    double *input = var % 2 == 0 ? params->data_gpu : params->output_gpu;
    double *output = var % 2 == 0 ? params->output_gpu : params->data_gpu;

    for (int i = 0; i < params->num_iter; ++i) {
        const int OFFSET = (i * params->num_vars + var) * (XG + 2) * (YG + 2) * (ZG + 2);
        double *read = input + OFFSET;
        double *write = output + OFFSET;
        cudaEventRecord(start);
        miniAMR << < grid, block >> > (
                read,
                write,
                gsz + 2,
                gsy + 2,
                gsx + 2,
                gsz + 2,
                gsy + 2,
                gsx + 2,
                gsx,
                gsy,
                gsz);
        cudaEventRecord(stop);

        CHECK_ERROR(cudaEventSynchronize(stop));
        float ms = 0.0;
        CHECK_ERROR(cudaEventElapsedTime(&ms, start, stop));
        kernel_event_time += (double) ms;
        if (ms > largest_diff_with_events) {
            largest_diff_with_events = ms;
        }

        num_kernel_calls += 1;
    }
#elif defined(V5)
    dim3 block(BLOCK_X, BLOCK_Y, BLOCK_Z);
    dim3 grid((gsx + BLOCK_X * EPT_X - 1) / (BLOCK_X * EPT_X),
            (gsy + BLOCK_Y * EPT_Y - 1) / (BLOCK_Y * EPT_Y),
            (gsz + BLOCK_Z * EPT_Z - 1) / (BLOCK_Z * EPT_Z));
    double *input = var % 2 == 0 ? params->data_gpu : params->output_gpu;
    double *output = var % 2 == 0 ? params->output_gpu : params->data_gpu;


    for (int i = 0; i < params->num_iter; ++i) {
        const int OFFSET = (i * params->num_vars + var) * (XG + 2) * (YG + 2) * (ZG + 2);
        double *read = input + OFFSET;
        double *write = output + OFFSET;
        CHECK_ERROR(ma_cudaMemcpyToSymbol(const_input, read, sizeof(double) * ((gsx + 2) * (gsy + 2) * (gsz + 2)), 0, cudaMemcpyDeviceToDevice));
        cudaEventRecord(start);
        miniAMR << < grid, block >> > (
                write,
                gsz + 2,
                gsy + 2,
                gsx + 2,
                gsz + 2,
                gsy + 2,
                gsx + 2,
                gsx,
                gsy,
                gsz);
        cudaEventRecord(stop);

        CHECK_ERROR(cudaEventSynchronize(stop));
        float ms = 0.0;
        CHECK_ERROR(cudaEventElapsedTime(&ms, start, stop));
        kernel_event_time += (double) ms;
        if (ms > largest_diff_with_events) {
            largest_diff_with_events = ms;
        }

        num_kernel_calls += 1;
    }

#elif defined(V6)

    dim3 block(BLOCK_X, BLOCK_Y, BLOCK_Z);
    dim3 grid((gsx + BLOCK_X * EPT_X - 1) / (BLOCK_X * EPT_X),
            (gsy + BLOCK_Y * EPT_Y - 1) / (BLOCK_Y * EPT_Y),
            (gsz + BLOCK_Z * EPT_Z - 1) / (BLOCK_Z * EPT_Z));
    double *input = var % 2 == 0 ? params->data_gpu : params->output_gpu;
    double *output = var % 2 == 0 ? params->output_gpu : params->data_gpu;

    // Since double is not supported in texture memory, we need to use int2 and cast it to double.
    // See: https://developer.nvidia.com/cuda-faq

    // auto channelDesc_input = cudaCreateChannelDesc<int2>();
    // cudaArray_t cuda_array;
    // CHECK_ERROR(cudaMalloc3DArray(&cuda_array, &channelDesc_input, make_cudaExtent((size_t) gsz + 2, (size_t) gsy + 2, (size_t) gsx + 2)));

    cudaTextureDesc input_texDesc;
    memset(&input_texDesc, 0, sizeof(cudaTextureDesc));
    input_texDesc.normalizedCoords = false;
    input_texDesc.filterMode = cudaFilterModePoint;
    input_texDesc.addressMode[0] = cudaAddressModeBorder;
    input_texDesc.addressMode[1] = cudaAddressModeBorder;
    input_texDesc.addressMode[2] = cudaAddressModeBorder;
    input_texDesc.readMode = cudaReadModeElementType;

    cudaResourceDesc input_resDesc;
    memset(&input_resDesc, 0, sizeof(cudaResourceDesc));
    input_resDesc.resType = cudaResourceTypeArray;
    input_resDesc.res.array.array = params->cuda_array;

    for (int i = 0; i < params->num_iter; ++i) {
        const int OFFSET = (i * params->num_vars + var) * (XG + 2) * (YG + 2) * (ZG + 2);
        double *read = input + OFFSET;
        double *write = output + OFFSET;

        cudaMemcpy3DParms input_parms = { 0 };
        input_parms.srcPtr = make_cudaPitchedPtr(read, sizeof(int2) * (gsz + 2), (size_t) gsz + 2, (size_t) gsy + 2);
        input_parms.dstArray = params->cuda_array;
        input_parms.extent = make_cudaExtent((size_t) gsz + 2, (size_t) gsy + 2, (size_t) gsx + 2);
        input_parms.kind = cudaMemcpyDeviceToDevice;
        CHECK_ERROR(ma_cudaMemcpy3D(&input_parms));

        cudaTextureObject_t input_tex;
        CHECK_ERROR(cudaCreateTextureObject(&input_tex, &input_resDesc, &input_texDesc, NULL));

        cudaEventRecord(start);
        miniAMR << < grid, block >> > (
                input_tex,
                write,
                0,
                0,
                0,
                gsz + 2,
                gsy + 2,
                gsx + 2,
                gsx,
                gsy,
                gsz);
        cudaEventRecord(stop);

        CHECK_ERROR(cudaDestroyTextureObject(input_tex));

        CHECK_ERROR(cudaEventSynchronize(stop));
        float ms = 0.0;
        CHECK_ERROR(cudaEventElapsedTime(&ms, start, stop));
        kernel_event_time += (double) ms;
        if (ms > largest_diff_with_events) {
            largest_diff_with_events = ms;
        }

        num_kernel_calls += 1;
    }
#else
#error "cuda version not defined"
#endif
    CHECK_ERROR(cudaEventDestroy(start));
    CHECK_ERROR(cudaEventDestroy(stop));
}

#if defined(V1) || defined(V2) || defined(V3) || defined(V4) || defined(V5) || defined(V6)
__global__ void
check_sum_kernel(int var, int num_iter, int num_vars, double *data, double *out_sum, int GS_X, int GS_Y, int GS_Z) {
    __shared__ double smem[BLOCK_X * BLOCK_Y * BLOCK_Z];

    int tid = (threadIdx.z * blockDim.y * blockDim.x) + (threadIdx.y * blockDim.x) + threadIdx.x;
    int idx = threadIdx.x + blockIdx.x * blockDim.x;
    int idy = threadIdx.y + blockIdx.y * blockDim.y;
    int idz = threadIdx.z + blockIdx.z * blockDim.z;

    smem[tid] = 0;
    constexpr int X = XG + 2;
    constexpr int Y = YG + 2;
    constexpr int Z = ZG + 2;

    if (!(idx >= GS_X || idy >= GS_Y || idz >= GS_Z)) {
        int idx1 = idx + 1;
        int idy1 = idy + 1;
        int idz1 = idz + 1;
        for (int i = 0; i < num_iter; ++i) {
            double *bp = &data[i * num_vars * X * Y * Z + var * X * Y * Z];
            smem[tid] += bp[idx1 * Y * Z + idy1 * Z + idz1];
        }
        __syncthreads();

        for (unsigned int s = 1; s < blockDim.x * blockDim.y * blockDim.z; s *= 2) {
            if (tid % (2 * s) == 0) {
                smem[tid] += smem[tid + s];
            }
            __syncthreads();
        }

        if (tid == 0) out_sum[blockIdx.z * gridDim.y * gridDim.x + blockIdx.y * gridDim.x + blockIdx.x] = smem[0];
    }
}

double check_sum_calc(int var, struct gpu_params *params) {
    // We initialize and de-initialize memory in the init_gpu and free_gpu methods.
    int gridSize_x = XG;
    int gridSize_y = YG;
    int gridSize_z = ZG;

    dim3 block(BLOCK_X, BLOCK_Y, BLOCK_Z);
    dim3 grid((gridSize_x + block.x * EPT_X - 1) / (block.x * EPT_X),
            (gridSize_y + block.y * EPT_Y - 1) / (block.y * EPT_Y),
            (gridSize_z + block.z * EPT_Z - 1) / (block.z * EPT_Z));

    double *sum_gpu;
    CHECK_ERROR(ma_cudaMalloc((void **) &sum_gpu, grid.x * grid.y * grid.z * sizeof(double)));

#if defined(V2) || defined(V3) || defined(V4)
    check_sum_kernel<<<grid, block>>>(
            var,
            params->num_iter,
            params->num_vars,
            (var - params->start) % 2 == 0 ? params->output_gpu : params->data_gpu,
            sum_gpu,
            gridSize_x,
            gridSize_y,
            gridSize_z);
#else
    check_sum_kernel<<<grid, block>>>(
            var,
            params->num_iter,
            params->num_vars,
            params->data_gpu,
            sum_gpu,
            gridSize_x,
            gridSize_y,
            gridSize_z);
#endif
    double sum[grid.x * grid.y * grid.z];
    CHECK_ERROR(ma_cudaMemcpy(&sum, sum_gpu, grid.x * grid.y * grid.z * sizeof(double), cudaMemcpyDeviceToHost));
    cudaFree(sum_gpu);

    for (int i = 1; i < grid.x * grid.y * grid.z; ++i) {
        sum[0] += sum[i];
    }
    return sum[0];
}

#elif defined(V0)
__global__ void
check_sum_kernel(int var, int num_iter, int num_vars, double *data, double *out_sum, int GS_X, int GS_Y) {
    __shared__ double smem[BLOCK_X * BLOCK_Y];

    int tid = threadIdx.y * blockDim.x + threadIdx.x;
    int idx = threadIdx.x + blockIdx.x * blockDim.x;
    int idy = threadIdx.y + blockIdx.y * blockDim.y;

    smem[tid] = 0;
    constexpr int X = XG + 2;
    constexpr int Y = YG + 2;
    constexpr int Z = ZG + 2;

    if (!(idx >= GS_X || idy >= GS_Y)) {
        int idx1 = idx + 1;
        int idy1 = idy + 1;
        for (int i = 0; i < num_iter; ++i) {
            double *bp = &data[i * num_vars * X * Y * Z + var * X * Y * Z];
            for (int z = 1; z <= ZG; ++z) {
                smem[tid] += bp[idx1 * Y * Z + idy1 * Z + z];
            }
        }
        __syncthreads();

        for (unsigned int s = 1; s < blockDim.x * blockDim.y; s *= 2) {
            if (tid % (2 * s) == 0) {
                smem[tid] += smem[tid + s];
            }
            __syncthreads();
        }

        if (tid == 0) out_sum[blockIdx.y * gridDim.x + blockIdx.x] = smem[0];
    }
}

double check_sum_calc(int var, struct gpu_params *params) {
    // We initialize and de-initialize memory in the init_gpu and free_gpu methods.
    int gridSize_x = XG;
    int gridSize_y = YG;

    dim3 block(BLOCK_X, BLOCK_Y);
    dim3 grid((gridSize_x
    +block.x * EPT_X - 1) / (block.x * EPT_X), (gridSize_y + block.y * EPT_Y - 1) / (block.y * EPT_Y));

    double *sum_gpu;
    CHECK_ERROR(ma_cudaMalloc((void **) &sum_gpu, grid.x * grid.y * sizeof(double)));

    check_sum_kernel << < grid, block >> >
                                (var, params->num_iter, params->num_vars, params->data_gpu, sum_gpu, gridSize_x, gridSize_y);

    double sum[grid.x * grid.y];
    CHECK_ERROR(ma_cudaMemcpy(&sum, sum_gpu, grid.x * grid.y * sizeof(double), cudaMemcpyDeviceToHost));
    cudaFree(sum_gpu);

    //printf("sum[0]: %f\n", sum[0]);
    for (int i = 1; i < grid.x * grid.y; ++i) {
        //printf("sum[%d]: %f\n", i, sum[i]);
        sum[0] += sum[i];
    }
    return sum[0];
}
#else
#error "cuda version not defined"
#endif
